import logging

from n2y.utils import fromisoformat

logger = logging.getLogger(__name__)


class Comment:
    """
    See https://developers.notion.com/reference/comment-object
    """

    def __init__(self, client, notion_data):
        self.notion_id = notion_data["id"]
        self.notion_parent = notion_data["parent"]
        self.discussion_id = notion_data["discussion_id"]
        self.created_time = fromisoformat(notion_data["created_time"])
        self.last_edited_time = fromisoformat(notion_data["last_edited_time"])
        self.created_by = client.wrap_notion_user(notion_data["created_by"])
        self.rich_text = client.wrap_notion_rich_text_array(
            notion_data["rich_text"], self
        )



import re
from collections import deque

from pandoc.types import (
    Str,
    Space,
    LineBreak,
    Strong,
    Emph,
    Strikeout,
    Code,
    Link,
    Underline,
    Math,
    InlineMath,
)

from n2y.logger import logger
from n2y.notion_mocks import mock_rich_text
from n2y.utils import pandoc_write_or_log_errors


class RichText:
    """
    A sequence of text that contains the same styling throughout. It also may
    contain an @-mention or an equation (both of which may be styled as well).

    Contains a reference to the block that the rich text is contained in, if
    any. If the block reference is None, then the rich text is perhaps used in
    a property_value or somewhere else.
    """

    def __init__(self, client, notion_data, block=None):
        self.client = client
        self.block = block

        self.plain_text = notion_data["plain_text"]
        self.href = notion_data.get("href", None)
        self.notion_type = notion_data["type"]
        self.notion_data = notion_data
        annotations = notion_data["annotations"]
        self.bold = annotations["bold"]
        self.italic = annotations["italic"]
        self.strikethrough = annotations["strikethrough"]
        self.underline = annotations["underline"]
        self.code = annotations["code"]
        self.color = annotations["color"]

    def to_pandoc(self):
        raise NotImplementedError()

    def to_value(self, pandoc_format):
        return pandoc_write_or_log_errors(
            self.to_pandoc(),
            format=pandoc_format,
            options=[],
        )

    @classmethod
    def plain_text_to_pandoc(cls, plain_text):
        ast = []
        match = re.findall(r"( +)|(\xa0+)|(\S+)|(\n+)|(\t+)", plain_text)
        for m in match:
            space, non_breaking_space, word, newline, tab = m
            for _ in range(len(space)):
                ast.append(Space())
            for _ in range(len(non_breaking_space)):
                ast.append(Space())
            if word:
                ast.append(Str(word))
            for _ in range(len(newline)):
                ast.append(LineBreak())
            for _ in range(len(tab) * 4):  # 4 spaces per tab
                ast.append(Space())
        return ast

    def annotate_pandoc_ast(self, target):
        """
        Pandoc's Strong, Emph, Underline, and Strikeout all accept sub-trees of
        inlines. Code does not. This is likely because `a *bold* word` does not
        create a bold word, but *a `code` word* does produce a bolded code word.
        Thus, it's not generally feasible for `annotate_pandoc_ast` to properly
        wrap any ast in a `Code`. If `Code` formatting is to be preserved, then
        the subclasses of `RichText` must apply it separately.
        """
        blank_space = [Space(), LineBreak()]

        if all(n in blank_space for n in target):
            return target

        prependages = deque()
        appendages = deque()
        problematic_annotations = [self.bold, self.italic, self.strikethrough]

        # Notion's data model allows space to be bolded, italicized, or
        # strick through, but markdown's doesn't. Thus, since these annotations
        # aren't visible around blank space anyway, we don't apply the
        # annotation to the blank space
        if any(problematic_annotations):
            while target[0] in blank_space:
                prependages.append(target.pop(0))
            while target[-1] in blank_space:
                appendages.appendleft(target.pop(-1))

        result = target
        if self.bold:
            result = [Strong(result)]
        if self.italic:
            result = [Emph(result)]
        if self.underline:
            result = [Underline(result)]
        if self.strikethrough:
            result = [Strikeout(result)]

        return list(prependages) + result + list(appendages)


class MentionRichText(RichText):
    def __init__(self, client, notion_data, block=None, mention=None):
        super().__init__(client, notion_data, block)
        if mention is None:
            self.mention = client.wrap_notion_mention(
                notion_data["mention"],
                notion_data["plain_text"],
                block,
            )
        else:
            self.mention = mention

    def to_pandoc(self):
        if self.code:
            logger.warning(
                'Code formatting is being dropped on mention "%s"', self.plain_text
            )
        mention_ast = self.mention.to_pandoc()
        return self.annotate_pandoc_ast(mention_ast)


class EquationRichText(RichText):
    def __init__(self, client, notion_data, block=None):
        super().__init__(client, notion_data, block)
        self.expression = notion_data["equation"]["expression"]

    def to_pandoc(self):
        if self.code:
            logger.warning(
                'Code formatting is being dropped on equation "%s"', self.expression
            )
        equation_ast = [Math(InlineMath(), self.expression)]
        return self.annotate_pandoc_ast(equation_ast)


class TextRichText(RichText):
    @classmethod
    def from_plain_text(cls, client, string, block=None):
        notion_data = mock_rich_text(string)
        return cls(client, notion_data, block)

    def __init__(self, client, notion_data, block=None):
        super().__init__(client, notion_data, block)

    def to_pandoc(self):
        if not self.code:
            plain_text_ast = self.plain_text_to_pandoc(self.plain_text)
            annotated_ast = self.annotate_pandoc_ast(plain_text_ast)
        else:
            code_ast = [Code(("", [], []), self.plain_text)]
            annotated_ast = self.annotate_pandoc_ast(code_ast)
        if self.href:
            return [Link(("", [], []), annotated_ast, (self.href, ""))]
        else:
            return annotated_ast


DEFAULT_RICH_TEXTS = {
    "text": TextRichText,
    "equation": EquationRichText,
    "mention": MentionRichText,
}


class RichTextArray:
    """
    Contains a sequence or styled text.

    For example, the sentence with a single bolded word

        "The man ran to *eat* the hot dog."

    would be represented as a `RichTextArray` with three elements. The first
    would contain "The man ran to ", the second would contain "eat", and would
    be bolded, and the third would contain " the hot dog.".
    """

    def __init__(self, client, notion_data, block=None):
        self.client = client
        self.block = block

        assert isinstance(notion_data, list)
        self.items = [client.wrap_notion_rich_text(i, block) for i in notion_data]

    def __len__(self):
        return len(self.items)

    def __iter__(self):
        return iter(self.items)

    def __getitem__(self, index):
        return self.items[index]

    def to_pandoc(self):
        return sum([item.to_pandoc() for item in self.items], [])

    def to_value(self, pandoc_format):
        return pandoc_write_or_log_errors(
            self.to_pandoc(),
            format=pandoc_format,
            options=[],
        )

    def to_plain_text(self):
        return "".join(item.plain_text for item in self.items)

    def matches(self, regexp):
        if len(self.items) > 0:
            first_item = self.items[0]

            # don't match EquationRichText or MentionRichText
            if isinstance(first_item, TextRichText):
                return regexp.match(first_item.plain_text)

    def lstrip(self, string):
        if len(self.items) > 0:
            first_item = self.items[0]
            if isinstance(first_item, TextRichText):
                string_len = len(string)
                if first_item.plain_text[:string_len] == string:
                    self.items[0].plain_text = first_item.plain_text[string_len:]
                    if self.items[0].plain_text == "":
                        self.items.pop(0)

    def prepend(self, string):
        rich_text = TextRichText.from_plain_text(self.client, string)
        self.items.insert(0, rich_text)



import pytest
from n2y.plugins.jinjarenderpage import FirstPassOutput


class TestFirstPassOutput:
    def test_create(self):
        first_pass_output = FirstPassOutput()
        assert not first_pass_output

    def test_first_pass_behavior(self):
        first_pass_output = FirstPassOutput()
        assert not first_pass_output.is_second_pass
        assert not first_pass_output.second_pass_is_requested
        assert first_pass_output.lines == []
        assert first_pass_output.second_pass_is_requested
        assert first_pass_output.lines == []
        assert first_pass_output.source == ""

    @pytest.mark.parametrize(
        "source_lines, expected_source",
        [
            (
                [],
                "",
            ),
            (
                ["alpha Beta", "GAMMA", "beta"],
                "alpha Beta\nGAMMA\nbeta",
            ),
        ],
    )
    def test_second_pass_behavior(self, source_lines, expected_source):
        first_pass_output = FirstPassOutput(source_lines)
        assert first_pass_output.is_second_pass
        assert not first_pass_output.second_pass_is_requested
        assert first_pass_output.lines == source_lines
        assert first_pass_output.source == expected_source
        assert not first_pass_output.second_pass_is_requested



"""
These tests verify how the n2y block classes convert notion data into Pandoc
abstract syntax tree (AST) objects, and then into markdown.
"""
from pandoc.types import Str, Space, Header, Link

from n2y.utils import strip_hyphens
from n2y.notion_mocks import mock_block, mock_rich_text
from tests.test_blocks import process_block

linked_headers = ["n2y.plugins.linkedheaders"]


def mock_header_ast(level, suffix, notion_block):
    return Header(
        level,
        ("", [], []),
        [
            Link(
                ("", [], []),
                [Str("Heading"), Space(), Str(suffix)],
                (f'#{strip_hyphens(notion_block["id"])}', ""),
            )
        ],
    )


def test_heading_1():
    notion_block = mock_block(
        "heading_1", {"rich_text": [mock_rich_text("Heading One")]}
    )
    pandoc_ast, markdown = process_block(notion_block, linked_headers)
    assert pandoc_ast == mock_header_ast(1, "One", notion_block)
    assert markdown == f'# [Heading One]({notion_block["url"]})\n'


def test_heading_1_bolding_stripped():
    notion_block = mock_block(
        "heading_1", {"rich_text": [mock_rich_text("Heading One", ["bold"])]}
    )
    pandoc_ast, markdown = process_block(notion_block, linked_headers)
    assert pandoc_ast == mock_header_ast(1, "One", notion_block)
    assert markdown == f'# [Heading One]({notion_block["url"]})\n'


def test_heading_2():
    notion_block = mock_block(
        "heading_2", {"rich_text": [mock_rich_text("Heading Two")]}
    )
    pandoc_ast, markdown = process_block(notion_block, linked_headers)
    assert pandoc_ast == mock_header_ast(2, "Two", notion_block)
    assert markdown == f'## [Heading Two]({notion_block["url"]})\n'


def test_heading_3():
    notion_block = mock_block(
        "heading_3", {"rich_text": [mock_rich_text("Heading Three")]}
    )
    pandoc_ast, markdown = process_block(notion_block, linked_headers)
    assert pandoc_ast == mock_header_ast(3, "Three", notion_block)
    assert markdown == f'### [Heading Three]({notion_block["url"]})\n'



from n2y.blocks import ChildDatabaseBlock, ChildPageBlock, TableOfContentsBlock
from n2y.logger import logger
from n2y.property_values import TitlePropertyValue
from n2y.utils import fromisoformat


class Page:
    def __init__(self, client, notion_data):
        logger.debug("Instantiating page")
        self.notion_parent = notion_data["parent"]
        self.archived = notion_data["archived"]
        self.notion_url = notion_data["url"]
        self.notion_id = notion_data["id"]
        self.notion_data = notion_data
        self.plugin_data = {}
        self._children = None
        self.client = client
        self._block = None

        self.created_time = fromisoformat(notion_data["created_time"])
        self.last_edited_time = fromisoformat(notion_data["last_edited_time"])
        self.created_by = client.wrap_notion_user(notion_data["created_by"])
        self.last_edited_by = client.wrap_notion_user(notion_data["last_edited_by"])
        self.icon = self._init_icon(notion_data["icon"])
        self.cover = notion_data["cover"] and client.wrap_notion_file(
            notion_data["cover"]
        )
        self.properties = {
            k: client.wrap_notion_property_value(npv, self)
            for k, npv in notion_data["properties"].items()
        }

    def _init_icon(self, icon_notion_data):
        """
        The icon property is unique in that it can be either an emoji or a file.
        """
        if icon_notion_data is None:
            return None
        elif icon_notion_data["type"] == "emoji":
            return self.client.wrap_notion_emoji(icon_notion_data)
        else:
            return self.client.wrap_notion_file(icon_notion_data)

    @property
    def title(self):
        for property_value in self.properties.values():
            # Notion ensures there is always exactly one title property
            if isinstance(property_value, TitlePropertyValue):
                return property_value.rich_text

    @property
    def block(self):
        if self._block is None:
            self._block = self.client.get_block(self.notion_id, page=self)
        return self._block

    @property
    def children(self):
        """
        Get a list of child pages and databases.
        """
        if self._children is None:
            self._children = []
            for block in self.block.children:
                self._append_children(block)
        return self._children

    def get_children(self):
        self.block.get_children()
        self._children = []
        for block in self.block.children:
            self._append_children(block)

    def _append_children(self, block):
        if isinstance(block, ChildPageBlock):
            page = self.client.get_page(block.notion_id)
            self._children.append(page)
        elif isinstance(block, ChildDatabaseBlock):
            database = self.client.get_database(block.notion_id)
            self._children.append(database)
        elif block.children is not None:
            # Recursively look for child pages and databases in the hierarchy
            for child_block in block.children:
                self._append_children(child_block)

    @property
    def parent(self):
        parent_type = self.notion_parent["type"]
        if parent_type == "workspace":
            return None
        elif parent_type == "page_id":
            return self.client.get_page(self.notion_parent["page_id"])
        else:
            assert parent_type == "database_id"
            return self.client.get_database(self.notion_parent["database_id"])

    def generate_toc(self, ast):
        """
        Generate a table of contents for the page.
        """
        toc_indecies = []

        def is_toc(i, block):
            if type(block) is TableOfContentsBlock:
                toc_indecies.append(i)
                return True
            return False

        if any(is_toc(*t) for t in enumerate(self.block.children)):
            child_ast = ast[1]
            for i in toc_indecies:
                self.block.children[i].render_toc(child_ast)
            ast = self.block.to_pandoc()
        return ast

    def to_pandoc(self, ignore_toc=False):
        ast = self.block.to_pandoc()
        return ast if ignore_toc else self.generate_toc(ast)

    def properties_to_values(self, pandoc_format=None):
        return {k: v.to_value(pandoc_format) for k, v in self.properties.items()}



import re

from n2y.blocks import HeadingThreeBlock
from n2y.errors import UseNextClass


class DeepHeadersBlock(HeadingThreeBlock):
    """
    Adds support for deeper headers in Notion (i.e., and h4, h5, h6, etc.).

    Any header 3 that begins with one or more "=" followed by a " " will be
    replaced with a deeper header. For example, a notion header 3 block that
    begins with "== Header" will become an h5 and the "== " will be stripped.
    """

    trigger_regex = re.compile(r"^(=+) ")

    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        result = self.rich_text.matches(self.trigger_regex)
        if result:
            number_of_equal_signs = len(result.group(1))
            self.level += number_of_equal_signs
            self.rich_text.lstrip(result.group(0))
        else:
            raise UseNextClass()


notion_classes = {
    "blocks": {
        "heading_3": DeepHeadersBlock,
    }
}



from itertools import groupby
from re import match
from urllib.parse import urljoin

from pandoc.types import (
    AlignDefault,
    BlockQuote,
    BulletList,
    Caption,
    Cell,
    CodeBlock,
    ColSpan,
    ColWidthDefault,
    Decimal,
    DisplayMath,
    Header,
    HorizontalRule,
    Image,
    Link,
    Math,
    OrderedList,
    Pandoc,
    Para,
    Period,
    Plain,
    Row,
    RowHeadColumns,
    RowSpan,
    Str,
    Table,
    TableBody,
    TableFoot,
    TableHead,
)

from n2y.logger import logger
from n2y.notion_mocks import mock_block, mock_rich_text_array
from n2y.utils import header_id_from_text, pandoc_write_or_log_errors, yaml_map_to_meta


class Block:
    """
    A Notion page's content consists of a tree of nested Blocks.

    See here for a listing of all of them: https://developers.notion.com/reference/block

    This class that wraps the Notion blocks and also is responsible for
    transforming them into Pandoc abstract syntax tree. A single Notion block
    often has multiple lines of text.

    Pandoc's data model doesn't line up exactly with Notion's, for example
    Notion doesn't have a wrapper around lists, while Pandoc does.
    """

    def __init__(self, client, notion_data, page=None, get_children=True):
        """
        The Notion client object is passed down for the following reasons:
        1. Some child objects may be unknown until the block is processed.
           Links to other Notion pages are an example.
        2. In some cases a block may choose not to get child blocks.
           Currently, all blocks load all children.
        """
        logger.debug('Instantiating "%s" block', type(self).__name__)
        self.client = client
        self.page = page

        self.notion_id = notion_data["id"]
        self.created_time = notion_data["created_time"]
        self.created_by = notion_data["created_by"]
        self.last_edited_time = notion_data["last_edited_time"]
        self.last_edited_by = notion_data["last_edited_by"]
        self.has_children = notion_data["has_children"]
        self.archived = notion_data["archived"]
        self.notion_type = notion_data["type"]
        self.notion_data = notion_data
        if get_children:
            self.get_children()
        else:
            self.children = None

    def get_children(self):
        if self.has_children:
            self.children = self.client.get_child_blocks(self.notion_id, self.page, True)
        else:
            self.children = []

    def to_pandoc(self):
        raise NotImplementedError()

    def children_to_pandoc(self):
        pandoc_ast = []
        for block_type, blocks in groupby(self.children, lambda c: type(c)):
            if issubclass(block_type, ListItemBlock):
                pandoc_ast.append(block_type.list_to_pandoc(blocks))
            else:
                for b in blocks:
                    result = b.to_pandoc()
                    if block_type == ChildPageBlock:
                        pandoc_ast.extend(result[1])
                    elif isinstance(result, list):
                        # a few blocks return lists of nodes
                        pandoc_ast.extend(result)
                    elif result is not None:
                        # a plugin may decide to return None to indicate the
                        # block should be removed; ideally pandoc.types.Nil
                        # would handle this, but it doesn't appear to work
                        pandoc_ast.append(result)
        return pandoc_ast

    @property
    def notion_type_data(self):
        return self.notion_data[self.notion_data["type"]]

    @property
    def notion_url(self):
        # the notion URL's don't work if the dashes from the block ID are present
        fragment = "#" + self.notion_id.replace("-", "")
        if self.page is None:
            return fragment
        else:
            return urljoin(self.page.notion_url, fragment)


class ListItemBlock(Block):
    @classmethod
    def list_to_pandoc(cls, items):
        raise NotImplementedError()


class ChildPageBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.title = self.notion_type_data["title"]

    def to_pandoc(self):
        assert self.children is not None
        children = self.children_to_pandoc()
        if self.page:
            properties = self.page.properties_to_values()
        else:
            properties = {}
        meta = yaml_map_to_meta(properties)
        return Pandoc(meta, children)


class EquationBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.expression = self.notion_type_data["expression"]

    def to_pandoc(self):
        return Para([Math(DisplayMath(), self.expression)])


class ParagraphBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.rich_text = client.wrap_notion_rich_text_array(
            self.notion_type_data["rich_text"], self
        )

    def to_pandoc(self):
        content = self.rich_text.to_pandoc()
        if self.has_children:
            # Notion allows you to create child blocks for a paragraph; these
            # child blocks appear indented relative to the paragraph. There's
            # no way to represent this indentation in pandoc's AST, so we just
            # append the child blocks afterwards.
            result = [Para(content)]
            children = self.children_to_pandoc()
            result.extend(children)
        else:
            result = Para(content)
        return result


class BulletedListItemBlock(ListItemBlock):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.rich_text = client.wrap_notion_rich_text_array(
            self.notion_type_data["rich_text"], self
        )

    def to_pandoc(self):
        content = [Plain(self.rich_text.to_pandoc())]
        if self.has_children:
            children = self.children_to_pandoc()
            for child in children:
                if isinstance(child, Table):
                    # A bug in pandoc's markdown reader makes it unable to register mid-list tables.
                    # This adds an extra space between the table and the previous list item, which
                    # Allows the table to be read as such.
                    content.append(Para([]))
                content.append(child)
        return content

    @classmethod
    def list_to_pandoc(cls, items):
        return BulletList([b.to_pandoc() for b in items])


class ToDoListItemBlock(BulletedListItemBlock):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.checked = self.notion_type_data["checked"]

        box = "☒ " if self.checked else "☐ "
        self.rich_text.prepend(box)


class NumberedListItemBlock(BulletedListItemBlock):
    @classmethod
    def list_to_pandoc(cls, items):
        return OrderedList((1, Decimal(), Period()), [b.to_pandoc() for b in items])


class TableOfContentsBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        self.subheaders: list[Header] | None = notion_data[notion_data["type"]].get(
            "subheaders", None
        )
        super().__init__(client, notion_data, page, get_children)

    def get_children(self):
        if self.subheaders is not None:
            children: list[TableOfContentsItemBlock] = []
            subsections: list[list[Header]] = []
            # Sometimes, the first header is not an H1, so we need to find the first
            base: int = self.subheaders[0][0]
            index: int = -1
            for header in self.subheaders:
                if header[0] == base:
                    index += 1
                    subsections.append([header])
                if header[0] > base:
                    subsections[index].append(header)
                if header[0] < base:
                    logger.warning(
                        f'Skipping out-of-order header "{header[1][0]}" in table of'
                        " contents for page named"
                        f" {self.page.title.to_plain_text()} ({self.page.notion_url})."
                        " The base header is an H{base} so all following headers"
                        " should be H{base} or greater."
                    )
            for subsection in subsections:
                notion_data = self.generate_item_block(subsection)
                children.append(
                    TableOfContentsItemBlock(self.client, notion_data, self.page)
                )
            if children:
                self.has_children = True
            self.children = children
        else:
            self.children = None

    def get_subheaders(self, ast_list):
        self.subheaders: list[Header] | None = []
        for block in ast_list:
            if isinstance(block, Header):
                self.subheaders.append(block)

    def to_pandoc(self):
        if not self.children:
            return None
        else:
            return self.children_to_pandoc()

    def generate_item_block(self, section: list[Header]):
        header = section.pop(0)
        header_text = pandoc_write_or_log_errors(
            header[2],
            format="gfm",
            options=[],
        )[:-1]
        rich_text = mock_rich_text_array([(header_text, None, f"#{header[1][0]}")])
        type_data = {
            "header": header,
            "subheaders": section,
            "rich_text": rich_text,
        }
        return mock_block("table_of_contents_item", type_data)

    def render_toc(self, ast_list):
        self.get_subheaders(ast_list)
        self.get_children()


class TableOfContentsItemBlock(NumberedListItemBlock, TableOfContentsBlock):
    def __init__(self, client, notion_data, page, get_children=True):
        type_data = notion_data[notion_data["type"]]
        self.header = type_data["header"]
        self.level = self.header[0]
        super().__init__(client, notion_data, page, get_children)

    def get_children(self):
        children: list[TableOfContentsItemBlock] = []
        subsections: list[list[Header]] = []
        index: int = -1
        for header in self.subheaders:
            if header[0] == self.level + 1:
                index += 1
                subsections.append([header])
            if header[0] > self.level + 1:
                try:
                    subsections[index].append(header)
                except IndexError:
                    logger.warning(
                        f'Skipping out-of-order header "{header[1][0]}" in table of'
                        " contents for page named"
                        f" {self.page.title.to_plain_text()} ({self.page.notion_url})."
                        " Please create headers in sequence (i.e. proper document"
                        f" formatting dictates that an H{self.level} should not be"
                        f" immediately followed by an H{header[0]}, but an"
                        f" H{self.level + 1})."
                    )
        for subsection in subsections:
            notion_data = self.generate_item_block(subsection)
            children.append(TableOfContentsItemBlock(self.client, notion_data, self.page))
        if children:
            self.has_children = True
        self.children = children


class HeadingBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.rich_text = client.wrap_notion_rich_text_array(
            self.notion_type_data["rich_text"], self
        )

        # The Notion UI allows one to bold the text in a header, but the bold
        # styling isn't displayed. Thus, to avoid unexpected appearances of
        # bold text in the generated documents, bolding is removed.
        for rich_text in self.rich_text:
            rich_text.bold = False

    def to_pandoc(self):
        section_id = header_id_from_text(self.rich_text.to_plain_text())
        return Header(self.level, (section_id, [], []), self.rich_text.to_pandoc())


class HeadingOneBlock(HeadingBlock):
    level = 1


class HeadingTwoBlock(HeadingBlock):
    level = 2


class HeadingThreeBlock(HeadingBlock):
    level = 3


class DividerBlock(Block):
    def to_pandoc(self):
        return HorizontalRule()


class BookmarkBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.url = self.notion_type_data["url"]
        self.caption = client.wrap_notion_rich_text_array(
            self.notion_type_data["caption"], self
        )

    def to_pandoc(self):
        if self.caption:
            caption_ast = self.caption.to_pandoc()
        else:
            caption_ast = [Str(self.url)]
        return Para([Link(("", [], []), caption_ast, (self.url, ""))])


class FencedCodeBlock(Block):
    pandoc_highlight_languages = [
        "abc",
        "actionscript",
        "ada",
        "agda",
        "apache",
        "asn1",
        "asp",
        "ats",
        "awk",
        "bash",
        "bibtex",
        "boo",
        "c",
        "changelog",
        "clojure",
        "cmake",
        "coffee",
        "coldfusion",
        "comments",
        "commonlisp",
        "cpp",
        "cs",
        "css",
        "curry",
        "d",
        "default",
        "diff",
        "djangotemplate",
        "dockerfile",
        "dot",
        "doxygen",
        "doxygenlua",
        "dtd",
        "eiffel",
        "elixir",
        "elm",
        "email",
        "erlang",
        "fasm",
        "fortranfixed",
        "fortranfree",
        "fsharp",
        "gcc",
        "glsl",
        "gnuassembler",
        "go",
        "graphql",
        "groovy",
        "hamlet",
        "haskell",
        "haxe",
        "html",
        "idris",
        "ini",
        "isocpp",
        "j",
        "java",
        "javadoc",
        "javascript",
        "javascriptreact",
        "json",
        "jsp",
        "julia",
        "kotlin",
        "latex",
        "lex",
        "lilypond",
        "literatecurry",
        "literatehaskell",
        "llvm",
        "lua",
        "m4",
        "makefile",
        "mandoc",
        "markdown",
        "mathematica",
        "matlab",
        "maxima",
        "mediawiki",
        "metafont",
        "mips",
        "modelines",
        "modula2",
        "modula3",
        "monobasic",
        "mustache",
        "nasm",
        "nim",
        "noweb",
        "objectivec",
        "objectivecpp",
        "ocaml",
        "octave",
        "opencl",
        "orgmode",
        "pascal",
        "perl",
        "php",
        "pike",
        "postscript",
        "povray",
        "powershell",
        "prolog",
        "protobuf",
        "pure",
        "purebasic",
        "python",
        "qml",
        "r",
        "raku",
        "relaxng",
        "relaxngcompact",
        "rest",
        "rhtml",
        "roff",
        "ruby",
        "rust",
        "sass",
        "scala",
        "scheme",
        "sci",
        "scss",
        "sed",
        "sgml",
        "sml",
        "spdxcomments",
        "sql",
        "sqlmysql",
        "sqlpostgresql",
        "stan",
        "stata",
        "swift",
        "systemverilog",
        "tcl",
        "tcsh",
        "texinfo",
        "toml",
        "typescript",
        "verilog",
        "vhdl",
        "xml",
        "xorg",
        "xslt",
        "xul",
        "yacc",
        "yaml",
        "zsh",
    ]

    # TODO: finish filling in this mapping from Notion language names to
    # pandoc's supported language names
    notion_to_pandoc_highlight_languages = {
        "c#": "cs",
        "c++": "cpp",
        "f#": "fsharp",
        "objective-c": "objectivec",
        "docker": "dockerfile",
        "coffee": "coffeescript",
        "shell": "bash",  # seems better than nothing
    }

    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.language = self.notion_type_data["language"]
        self.rich_text = client.wrap_notion_rich_text_array(
            self.notion_type_data["rich_text"], self
        )
        self.caption = client.wrap_notion_rich_text_array(
            self.notion_type_data["caption"], self
        )

    def to_pandoc(self):
        pandoc_language = self.notion_to_pandoc_highlight_languages.get(
            self.language,
            self.language,
        )
        if pandoc_language not in self.pandoc_highlight_languages:
            if pandoc_language != "plain text":
                msg = 'Dropping syntax highlighting for unsupported language "%s" (%s)'
                logger.warning(msg, pandoc_language, self.notion_url)
            language = []
        else:
            language = [pandoc_language]
        return CodeBlock(("", language, []), self.rich_text.to_plain_text())


class QuoteBlock(ParagraphBlock):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.rich_text = client.wrap_notion_rich_text_array(
            self.notion_type_data["rich_text"], self
        )

    def to_pandoc(self):
        pandoc_ast = super().to_pandoc()
        return (
            BlockQuote(pandoc_ast)
            if isinstance(pandoc_ast, list)
            else BlockQuote([pandoc_ast])
        )


class FileBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.file = client.wrap_notion_file(notion_data["file"])
        self.caption = client.wrap_notion_rich_text_array(
            self.notion_type_data["caption"], self
        )
        potential_name = match(
            (
                r".+(?:[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}/)+"
                r"(?P<name>.+)\?.+"
            ),
            self.file.url,
        )
        name = potential_name.groups("name") if potential_name else None
        self.name = name[0] if name else None

    def to_pandoc(self):
        url = None
        if self.file.type == "external":
            url = self.file.url
        elif self.file.type == "file":
            url = self.client.download_file(self.file.url, self.page, self.notion_id)
        content_ast = [Link(("", [], []), [Str(self.name or url)], (url, ""))]
        if self.caption:
            caption_ast = self.caption.to_pandoc()
            return render_with_caption(content_ast, caption_ast)
        return Para(content_ast)


class ImageBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.file = client.wrap_notion_file(notion_data["image"])
        self.caption = client.wrap_notion_rich_text_array(
            self.notion_type_data["caption"], self
        )

    def to_pandoc(self):
        url = None
        if self.file.type == "external":
            url = self.file.url
        elif self.file.type == "file":
            url = self.client.download_file(self.file.url, self.page, self.notion_id)
        caption = []
        fig_flag = ""
        if self.caption:
            fig_flag = "fig:"
            caption = self.caption.to_pandoc()
        return Para([Image(("", [], []), caption, (url, fig_flag))])


class TableBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.has_column_header = self.notion_type_data["has_column_header"]
        self.has_row_header = self.notion_type_data["has_row_header"]
        self.table_width = self.notion_type_data["table_width"]

    def to_pandoc(self):
        children = self.children_to_pandoc()
        # Isolate the header row if it exists
        if self.has_column_header:
            header_rows = [children.pop(0)]
        else:
            header_rows = []
        if self.has_row_header:
            row_header_columns = 1
        else:
            row_header_columns = 0
        # Notion does not have cell alignment or width options, sticking with defaults.
        colspec = [(AlignDefault(), ColWidthDefault()) for _ in range(self.table_width)]
        table = Table(
            ("", [], []),
            Caption(None, []),
            colspec,
            TableHead(("", [], []), header_rows),
            [TableBody(("", [], []), RowHeadColumns(row_header_columns), [], children)],
            TableFoot(("", [], []), []),
        )
        return table


class RowBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.cells = [
            client.wrap_notion_rich_text_array(nc, self)
            for nc in self.notion_type_data["cells"]
        ]

    def to_pandoc(self):
        cells = []
        for cell in self.cells:
            pandoc = cell.to_pandoc()
            cells.append(
                Cell(
                    ("", [], []),
                    AlignDefault(),
                    RowSpan(1),
                    ColSpan(1),
                    [Plain(pandoc)],
                )
            )
        return Row(("", [], []), cells)


class ColumnListBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)

    def to_pandoc(self):
        pandoc = []
        if self.children:
            pandoc = self.children_to_pandoc()
        return pandoc


class ColumnBlock(Block):
    def to_pandoc(self):
        return self.children_to_pandoc()


class ToggleBlock(Block):
    """
    Generates a bulleted list item with indented children. A plugin may be used
    to add html classes and replicate the interactive behavior found in Notion.
    """

    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.rich_text = client.wrap_notion_rich_text_array(
            self.notion_type_data["rich_text"], self
        )

    def to_pandoc(self):
        header = self.rich_text.to_pandoc()
        children = self.children_to_pandoc()
        content = [Para(header)]
        content.extend(children)
        return BulletList([content])


class CalloutBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.rich_text = client.wrap_notion_rich_text_array(
            self.notion_type_data["rich_text"], self
        )
        # the color and icon are not currently used

    def to_pandoc(self):
        content = self.rich_text.to_pandoc()
        if self.has_children:
            children = self.children_to_pandoc()
            result = [Para(content)]
            result.extend(children)
        else:
            result = Para(content)
        return result


class NoopBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        # don't get the child blocks, as we're not using the data
        super().__init__(client, notion_data, page, get_children=False)

    def to_pandoc(self):
        return None


class BreadcrumbBlock(NoopBlock):
    pass


class UnsupportedBlock(NoopBlock):
    pass


class TemplateBlock(NoopBlock):
    pass


class WarningBlock(NoopBlock):
    def to_pandoc(self):
        logger.warning(
            'Skipping unsupported "%s" block (%s)', self.notion_type, self.notion_url
        )
        return None


class ChildDatabaseBlock(NoopBlock):
    pass


class EmbedBlock(WarningBlock):
    pass


class ContentBlock(Block):
    """
    Generic base class for blocks that contain file-based content that should be
    downloaded and then linked to.
    """

    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.file = client.wrap_notion_file(self.notion_type_data)
        self.caption = client.wrap_notion_rich_text_array(
            self.notion_type_data["caption"], self
        )

    def to_pandoc(self):
        url = None
        if self.file.type == "external":
            url = self.file.url
        elif self.file.type == "file":
            url = self.client.download_file(self.file.url, self.page, self.notion_id)
        content_ast = [Link(("", [], []), [Str(url)], (url, ""))]
        if self.caption:
            caption_ast = self.caption.to_pandoc()
            return render_with_caption(content_ast, caption_ast)
        else:
            return Para(content_ast)


class AudioBlock(ContentBlock):
    pass


class VideoBlock(ContentBlock):
    pass


class PdfBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.pdf = client.wrap_notion_file(notion_data["pdf"])
        self.caption = client.wrap_notion_rich_text_array(
            self.notion_type_data["caption"], self
        )

    def to_pandoc(self):
        url = self.client.download_file(self.pdf.url, self.page, self.notion_id)
        content_ast = [Link(("", [], []), [Str(url)], (url, ""))]
        if self.caption:
            caption_ast = self.caption.to_pandoc()
            return render_with_caption(content_ast, caption_ast)
        return Para(content_ast)


class ChildrenPassThroughBlock(Block):
    """
    Just passes along the children of the block.
    """

    def to_pandoc(self):
        return self.children_to_pandoc()


class LinkPreviewBlock(WarningBlock):
    pass


class SyncedBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        self.original = notion_data[notion_data["type"]]["synced_from"] is None
        super().__init__(client, notion_data, page, get_children=self.original)
        self.is_recursive = None
        # Synced blocks will always have children unless not shared
        # (There will always be at least one UnsupportedBlock child)
        self.shared = self.has_children
        self.children = self._get_synced_block_children()

    def _get_synced_block_children(self):
        if not self.original and self.shared:
            # This last condition is to protect against recursive synced blocks while
            # still allowing synced blocks that are children of other synced blocks
            # (once Notion confirms that this bug has been addressed it can be removed)
            parent = self.notion_data.get("parent", None)
            self.is_recursive = (
                parent
                and self.notion_type_data["synced_from"]["block_id"]
                == parent[parent["type"]]
            )
            if not self.is_recursive:
                return self.client.get_child_blocks(
                    self.notion_type_data["synced_from"]["block_id"],
                    self.page,
                    True,
                )
        return self.children

    def to_pandoc(self):
        if not self.shared:
            # logger.warning('Skipping un-shared synced block (%s)', self.notion_url)
            return None
        elif self.is_recursive:
            logger.warning("Skipping recursive synced block (%s)", self.notion_url)
            return None
        return self.children_to_pandoc()


class LinkToPageBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)

        self.link_type = self.notion_type_data["type"]
        # The key for the object id may be either "page_id"
        # or "database_id".
        self.linked_node_id = self.notion_type_data[self.link_type]

    def to_pandoc(self):
        # TODO: in the future, if we are exporting the linked page too, then add
        # a link to the page. For now, we just display the text of the page.
        if self.link_type == "page_id":
            node = self.client.get_page(self.linked_node_id)
        elif self.link_type == "database_id":
            node = self.client.get_database(self.linked_node_id)
        else:
            raise NotImplementedError(f"Unknown link type: {self.link_type}")

        if node is None:
            msg = "Permission denied when attempting to access linked node (%r)"
            logger.warning(msg, self.notion_url)
            return None
        else:
            title = node.title.to_pandoc()
            return Para(title)


def render_with_caption(content_ast, caption_ast):
    header_cell_args = [
        ("", [], []),
        AlignDefault(),
        RowSpan(1),
        ColSpan(1),
        [Plain(content_ast)],
    ]
    body_cell_args = [
        ("", [], []),
        AlignDefault(),
        RowSpan(1),
        ColSpan(1),
        [Plain(caption_ast)],
    ]
    body_row = Row(("", [], []), [Cell(*body_cell_args)])
    return Table(
        ("", [], []),
        Caption(None, []),
        [(AlignDefault(), ColWidthDefault())],
        TableHead(("", [], []), [Row(("", [], []), [Cell(*header_cell_args)])]),
        [TableBody(("", [], []), RowHeadColumns(0), [], [body_row])],
        TableFoot(("", [], []), []),
    )


DEFAULT_BLOCKS = {
    "paragraph": ParagraphBlock,
    "heading_1": HeadingOneBlock,
    "heading_2": HeadingTwoBlock,
    "heading_3": HeadingThreeBlock,
    "bulleted_list_item": BulletedListItemBlock,
    "numbered_list_item": NumberedListItemBlock,
    "to_do": ToDoListItemBlock,
    "toggle": ToggleBlock,
    "child_page": ChildPageBlock,
    "child_database": ChildDatabaseBlock,
    "embed": EmbedBlock,
    "image": ImageBlock,
    "audio": AudioBlock,
    "video": VideoBlock,
    "file": FileBlock,
    "pdf": PdfBlock,
    "bookmark": BookmarkBlock,
    "callout": CalloutBlock,
    "quote": QuoteBlock,
    "equation": EquationBlock,
    "divider": DividerBlock,
    "table_of_contents": TableOfContentsBlock,
    "table_of_contents_item": TableOfContentsItemBlock,
    "breadcrumb": BreadcrumbBlock,
    "column": ColumnBlock,
    "column_list": ColumnListBlock,
    "link_preview": LinkPreviewBlock,
    "synced_block": SyncedBlock,
    "template": TemplateBlock,
    "link_to_page": LinkToPageBlock,
    "code": FencedCodeBlock,
    "table": TableBlock,
    "table_row": RowBlock,
    "unsupported": UnsupportedBlock,
}



"""
This module contains all the code responsible for exporting `page.Page` and
`database.Database` objects into the various supported file formats.
"""
import os

import yaml
from pandoc.types import Table

from n2y.logger import logger
from n2y.utils import (
    pandoc_format_to_file_extension,
    pandoc_write_or_log_errors,
    sanitize_filename,
)


def _page_properties(
    page,
    pandoc_format=None,
    id_property=None,
    url_property=None,
    property_map=None,
):
    if pandoc_format is None:
        pandoc_format = "markdown"
    if property_map is None:
        property_map = {}
    properties = page.properties_to_values(pandoc_format)
    if id_property in properties:
        logger.warning(
            'The id property "%s" is shadowing an existing property with the same name',
            id_property,
        )
    if id_property:
        properties[id_property] = page.notion_id

    if url_property in properties:
        logger.warning(
            'The url property "%s" is shadowing an existing property with the same name',
            url_property,
        )
    if url_property:
        properties[url_property] = page.notion_url
    for original, new in property_map.items():
        if original in properties:
            value = properties.pop(original)
            if new:
                properties[new] = value
        else:
            msg = "Property %s not found in page %s; skipping remapping from %s to %s"
            logger.warning(msg, original, page.notion_url, original, new)
    return properties


def export_page(
    page,
    pandoc_format,
    pandoc_options,
    yaml_front_matter=True,
    id_property=None,
    url_property=None,
    property_map=None,
):
    pandoc_ast = page.to_pandoc()

    if (number_empty_headers := _count_headerless_tables(pandoc_ast)) > 0:
        if "markdown" in pandoc_format or "gfm" in pandoc_format:
            logger.warning(
                "%d table(s) will present empty headers to maintain Markdown spec (%r)",
                number_empty_headers,
                page.notion_url,
            )

    page_content = pandoc_write_or_log_errors(pandoc_ast, pandoc_format, pandoc_options)
    if isinstance(page_content, str) and yaml_front_matter:
        page_properties = _page_properties(
            page,
            pandoc_format,
            id_property,
            url_property,
            property_map,
        )
        return "\n".join(
            [
                "---",
                yaml.dump(page_properties) + "---",
                page_content,
            ]
        )
    else:
        # if the result is a binary file, return it as is (since we can't add YAML metadata to it)
        return page_content


def _count_headerless_tables(pandoc_ast):
    """
    Count the number of tables in the AST that will result in empty
    header rows prepended by Pandoc.
    """
    number_empty_headers = 0
    if pandoc_ast and any(isinstance(e, Table) for e in pandoc_ast[1]):
        for element in pandoc_ast[1]:
            if isinstance(element, Table):
                _, head = element[3]
                # We do count empty rows
                number_header_rows = sum(1 for _, row in head)
                if number_header_rows == 0:
                    number_empty_headers += 1
    return number_empty_headers


def database_to_yaml(
    database,
    pandoc_format,
    pandoc_options,
    notion_filter=None,
    notion_sorts=None,
    id_property=None,
    url_property=None,
    content_property=None,
    property_map=None,
):
    if content_property in database.schema:
        logger.warning(
            (
                'The content property "%s" is shadowing an existing '
                "property with the same name"
            ),
            content_property,
        )
    results = []
    for page in database.children_filtered(notion_filter, notion_sorts):
        result = _page_properties(
            page, pandoc_format, id_property, url_property, property_map
        )
        if content_property:
            pandoc_ast = page.to_pandoc()
            if pandoc_ast:
                result[content_property] = pandoc_write_or_log_errors(
                    pandoc_ast,
                    pandoc_format,
                    pandoc_options,
                )
            else:
                result[content_property] = None
        results.append(result)
    return results


def database_to_files(
    database,
    directory,
    pandoc_format,
    pandoc_options,
    yaml_front_matter=True,
    filename_template=None,
    notion_filter=None,
    notion_sorts=None,
    id_property=None,
    url_property=None,
    property_map=None,
):
    seen_file_names = set()
    counts = {"unnamed": 0, "duplicate": 0}
    for page in database.children_filtered(notion_filter, notion_sorts):
        page_filename = _page_filename(page, pandoc_format, filename_template)
        if page_filename:
            if page_filename not in seen_file_names:
                seen_file_names.add(page_filename)
                document = export_page(
                    page,
                    pandoc_format,
                    pandoc_options,
                    yaml_front_matter,
                    id_property,
                    url_property,
                    property_map,
                )
                write_document(document, os.path.join(directory, page_filename))
            else:
                logger.warning(
                    'Skipping page named "%s" since it has been used', page_filename
                )
                counts["duplicate"] += 1
        else:
            counts["unnamed"] += 1
    for key, count in counts.items():
        if count > 0:
            logger.info("%d %s page(s) skipped", count, key)


def write_document(document, path):
    if isinstance(document, bytes):
        file_mode = "wb"
    else:
        file_mode = "w"
    if os.path.dirname(path):
        os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, file_mode) as f:
        f.write(document)


def _page_filename(page, pandoc_format, filename_template=None):
    page_title = page.title.to_plain_text()
    if filename_template is None:
        extension = pandoc_format_to_file_extension(pandoc_format)
        return sanitize_filename(f"{page_title}.{extension}")
    else:
        page_properties = page.properties_to_values()
        if "TITLE" not in page_properties:
            page_properties["TITLE"] = page_title
        try:
            return sanitize_filename(filename_template.format(**page_properties))
        except KeyError:
            logger.warning(
                'Invalid filename property, "%s". Valid options are %s',
                filename_template,
                ", ".join(page.properties.keys()),
            )
            return _page_filename(page, pandoc_format)



import logging


logger = logging.getLogger(__name__)


class Emoji:
    """
    See https://developers.notion.com/reference/emoji-object
    """

    def __init__(self, client, notion_data):
        self.client = client
        self.emoji = notion_data["emoji"]



import re

from pandoc.types import Note, Str, Para

from n2y.logger import logger
from n2y.errors import UseNextClass
from n2y.blocks import ParagraphBlock
from n2y.rich_text import TextRichText


plugin_data_key = "n2y.plugins.footnotes"


class ParagraphWithFootnoteBlock(ParagraphBlock):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        if self._is_footnote():
            self._attach_footnote_data()
        else:
            raise UseNextClass()

    def to_pandoc(self):
        return None

    def _attach_footnote_data(self):
        if plugin_data_key not in self.page.plugin_data:
            self.page.plugin_data[plugin_data_key] = {}
        if self._footnote() not in self.page.plugin_data[plugin_data_key]:
            self.page.plugin_data[plugin_data_key][
                self._footnote()
            ] = self._footnote_ast()
            if self._footnote_empty():
                msg = 'Empty footnote "[%s]" (%s)'
                logger.warning(msg, self._footnote(), self.notion_url)
        else:
            msg = 'Multiple footnotes for "[%s]", skipping latest (%s)'
            logger.warning(msg, self._footnote(), self.notion_url)

    def _is_footnote(self):
        return self._footnote() is not None

    def _footnote(self):
        first_str = self.rich_text.to_plain_text().split(" ")[0]
        footnotes = re.findall(r"\[(\d+)\]:", first_str)
        if len(footnotes) != 1:
            return None
        return footnotes[0]

    def _footnote_ast(self):
        ast = super().to_pandoc()
        if isinstance(ast, list):
            first_paragraph_footnote_stripped = Para(ast[0][0][2:])
            remaining_paragraphs = ast[1:]
            return [first_paragraph_footnote_stripped] + remaining_paragraphs
        else:
            paragraph_footnote_stripped = Para(ast[0][2:])
            return paragraph_footnote_stripped

    def _footnote_empty(self):
        return len(self.rich_text.to_plain_text()) == 0


class TextRichTextWithFootnoteRef(TextRichText):
    def __init__(self, client, notion_data, block=None):
        super().__init__(client, notion_data, block)
        if not self._is_footnote():
            raise UseNextClass()

    def to_pandoc(self):
        pandoc_ast = []
        for token in super().to_pandoc():
            ref = self._footnote_from_token(token)
            if ref is None:
                pandoc_ast.append(token)
                continue
            if ref not in self.block.page.plugin_data[plugin_data_key]:
                pandoc_ast.append(token)
                msg = 'Missing footnote "[%s]". Rendering as plain text (%s)'
                logger.warning(msg, ref, self.block.notion_url)
                continue
            self._append_footnote_to_ast(pandoc_ast, token, ref)
        return pandoc_ast

    def _append_footnote_to_ast(self, pandoc_ast, token, ref):
        block = self.block.page.plugin_data[plugin_data_key][ref]
        footnote = Note(block) if isinstance(block, list) else Note([block])
        prefix, suffix = token[0].split(f"[^{ref}]")
        pandoc_ast.append(Str(prefix))
        pandoc_ast.append(footnote)
        pandoc_ast.append(Str(suffix))

    def _is_footnote(self):
        return any(
            self._footnote_from_token(t) is not None for t in super().to_pandoc()
        )

    def _footnote_from_token(self, token):
        if not isinstance(token, Str):
            return None
        refs = re.findall(r"\[\^(\d+)\]", token[0])
        if len(refs) != 1:
            return None
        return refs[0]


notion_classes = {
    "blocks": {"paragraph": ParagraphWithFootnoteBlock},
    "rich_texts": {"text": TextRichTextWithFootnoteRef},
}



from unittest.mock import PropertyMock, patch

from jinja2 import TemplateSyntaxError
from pytest import raises

from n2y.blocks import ChildPageBlock
from n2y.notion import Client
from n2y.notion_mocks import (
    mock_block,
    mock_database,
    mock_database_mention,
    mock_page,
    mock_rich_text,
    mock_rich_text_array,
    mock_user,
)
from n2y.page import Page
from n2y.plugins.jinjarenderpage import (
    JinjaFencedCodeBlock,
    JinjaRenderPage,
    fuzzy_find_in,
    join_to,
    list_matches,
    render_from_string,
)
from n2y.user import User
from n2y.utils import pandoc_ast_to_markdown, pandoc_write_or_log_errors

format_def = "{jinja=gfm}"
spaced_format_def = "{jinja=gfm} "


def test_join_to_basic():
    foreign_keys = ["1", "3"]
    table = [
        {"notion_id": "1", "data": "a"},
        {"notion_id": "2", "data": "b"},
    ]
    assert join_to(foreign_keys, table) == [{"notion_id": "1", "data": "a"}, None]
    assert join_to(foreign_keys, table, "data") == [None, None]


def test_fuzzy_find_in():
    dict_list = [
        {"id": "1", "data": "a"},
        {"id": "2", "data": "bc"},
        {"id": "3", "data": "def"},
    ]
    reversed_dict_list = [
        {"id": "3", "data": "def"},
        {"id": "2", "data": "bc"},
        {"id": "1", "data": "a"},
    ]
    catch_all_string = "a bc def"
    assert fuzzy_find_in(dict_list, "a", "data") == [{"id": "1", "data": "a"}]
    assert fuzzy_find_in(dict_list, catch_all_string, "data") == reversed_dict_list
    assert fuzzy_find_in(dict_list, catch_all_string, "data", False) == reversed_dict_list
    assert fuzzy_find_in(dict_list, catch_all_string, "data", False, False) == dict_list
    assert fuzzy_find_in(dict_list, catch_all_string, "data", True, False) == dict_list


def test_render_no_filtering():
    input_string = "apple\nbanana\ncherry\n"
    expected_result = input_string
    actual_result = render_from_string(input_string)
    assert actual_result == expected_result


def test_undefined():
    with raises(TemplateSyntaxError):
        input_string = "{% huhwhat 'hotel', 'california' %}"
        render_from_string(input_string)


@patch("n2y.notion.Client.wrap_notion_user")
def process_jinja_block(client, caption, jinja_code, wrap_notion_user):
    wrap_notion_user.return_value = User(client, mock_user())
    page_notion_data = mock_page()
    page = JinjaRenderPage(client, page_notion_data)

    page_block_notion_data = mock_block("child_page", {"title": "Mock Page"})
    page_block = ChildPageBlock(client, page_block_notion_data, page, False)

    page._block = page_block

    code_block_notion_data = mock_block(
        "code",
        {
            "language": "plain text",
            "caption": caption,
            "rich_text": mock_rich_text_array(jinja_code),
        },
    )
    code_block = JinjaFencedCodeBlock(client, code_block_notion_data, page, False)
    page.block.children = [code_block]
    return page


def test_jinja_syntax_err():
    client = Client("")
    caption = mock_rich_text_array(format_def)
    jinja_code = "{% huhwhat 'hotel', 'california' %}"
    page = process_jinja_block(client, caption, jinja_code)
    pandoc_ast = page.to_pandoc()
    markdown = pandoc_ast_to_markdown(pandoc_ast)
    assert "syntax error" in markdown
    assert "Encountered unknown tag 'huhwhat'" in markdown


def test_jinja_render_gfm():
    client = Client("")
    caption = mock_rich_text_array(format_def)
    jinja_code = "{% for v in ['a', 'b'] %}{{v}}{% endfor %}"
    page = process_jinja_block(client, caption, jinja_code)
    pandoc_ast = page.to_pandoc()
    markdown = pandoc_ast_to_markdown(pandoc_ast)
    assert markdown == "ab\n"


def test_jinja_render_gfm_with_second_pass():
    client = Client("")
    caption = mock_rich_text_array(format_def)
    jinja_code = "a{% for v in first_pass_output.lines %}{{v}}{% endfor %}"
    page = process_jinja_block(client, caption, jinja_code)
    page_block = page.block
    with patch.object(Page, "block", new_callable=PropertyMock) as mock_block:
        mock_block.return_value = page_block
        pandoc_ast = page.to_pandoc()
    markdown = pandoc_ast_to_markdown(pandoc_ast)
    assert markdown == "aa\n"


def test_jinja_render_html():
    client = Client("")
    caption = mock_rich_text_array("{jinja=html}")
    jinja_code = (
        "<table>"
        "<tr><th>Name</th></tr>"
        "{% for v in ['a', 'b'] %}<tr><td>{{v}}</td></tr>{% endfor %}"
        "</table>"
    )
    page = process_jinja_block(client, caption, jinja_code)
    pandoc_ast = page.to_pandoc()
    markdown = pandoc_ast_to_markdown(pandoc_ast)
    assert markdown == "  Name\n  ------\n  a\n  b\n"


@patch("n2y.notion.Client.wrap_notion_user")
def test_jinja_render_with_database(wrap_notion_user):
    client = Client("")
    wrap_notion_user.return_value = User(client, mock_user())
    database_notion_data = mock_database(title="My DB")
    mention_notion_data = mock_database_mention(database_notion_data["id"])
    database_pages_notion_data = [mock_page(title="a"), mock_page(title="b")]
    caption = [
        mock_rich_text(spaced_format_def),
        mock_rich_text("My DB", mention=mention_notion_data),
    ]
    jinja_code = "{% for v in databases['My DB'] %}{{v.title}}{% endfor %}"

    with patch.object(Client, "get_notion_database") as mock_get_notion_database:
        with patch.object(
            Client, "get_database_notion_pages"
        ) as mock_get_database_notion_pages:
            mock_get_notion_database.return_value = database_notion_data
            mock_get_database_notion_pages.return_value = database_pages_notion_data
            page = process_jinja_block(client, caption, jinja_code)
            pandoc_ast = page.to_pandoc()

    markdown = pandoc_ast_to_markdown(pandoc_ast)
    assert markdown == "ab\n"


@patch("n2y.notion.Client.wrap_notion_user")
def test_jinja_render_with_missing_database(wrap_notion_user):
    client = Client("")
    wrap_notion_user.return_value = User(client, mock_user())
    database_notion_data = mock_database(title="My DB")
    mention_notion_data = mock_database_mention(database_notion_data["id"])
    database_pages_notion_data = [mock_page(title="a"), mock_page(title="b")]
    caption = [
        mock_rich_text(spaced_format_def),
        mock_rich_text("My DB", mention=mention_notion_data),
    ]
    jinja_code = "{{ databases['MISSING'] }}"

    with patch.object(Client, "get_notion_database") as mock_get_notion_database:
        with patch.object(
            Client, "get_database_notion_pages"
        ) as mock_get_database_notion_pages:
            mock_get_notion_database.return_value = database_notion_data
            mock_get_database_notion_pages.return_value = database_pages_notion_data
            page = process_jinja_block(client, caption, jinja_code)
            pandoc_ast = page.to_pandoc()

    markdown = pandoc_ast_to_markdown(pandoc_ast)
    assert 'You attempted to access the "MISSING" database' in markdown
    assert 'the only available database is "My DB".' in markdown


@patch("n2y.notion.Client.wrap_notion_user")
def test_jinja_render_with_incorrect_db_property(wrap_notion_user):
    client = Client("")
    wrap_notion_user.return_value = User(client, mock_user())
    database_notion_data = mock_database(title="My DB")
    mention_notion_data = mock_database_mention(database_notion_data["id"])
    database_pages_notion_data = [mock_page(title="a"), mock_page(title="b")]
    caption = [
        mock_rich_text(spaced_format_def),
        mock_rich_text("My DB", mention=mention_notion_data),
    ]
    jinja_code = "{{ databases[0][0].Foo }}"

    with patch.object(Client, "get_notion_database") as mock_get_notion_database:
        with patch.object(
            Client, "get_database_notion_pages"
        ) as mock_get_database_notion_pages:
            mock_get_notion_database.return_value = database_notion_data
            mock_get_database_notion_pages.return_value = database_pages_notion_data
            page = process_jinja_block(client, caption, jinja_code)
            pandoc_ast = page.to_pandoc()

    markdown = pandoc_ast_to_markdown(pandoc_ast)
    assert (
        'You attempted to access the "Foo" property of a database item on line 1'
        in markdown
    )
    assert (
        'the available properties are "title", "notion_id", and "notion_url".' in markdown
    )


@patch("n2y.notion.Client.wrap_notion_user")
def test_jinja_render_with_missing_page_property(wrap_notion_user):
    client = Client("")
    wrap_notion_user.return_value = User(client, mock_user())
    database_notion_data = mock_database(title="My DB")
    mention_notion_data = mock_database_mention(database_notion_data["id"])
    database_pages_notion_data = [mock_page(title="a"), mock_page(title="b")]
    caption = [
        mock_rich_text(spaced_format_def),
        mock_rich_text("My DB", mention=mention_notion_data),
    ]
    jinja_code = "{{ page['MISSING'] }}"

    with patch.object(Client, "get_notion_database") as mock_get_notion_database:
        with patch.object(
            Client, "get_database_notion_pages"
        ) as mock_get_database_notion_pages:
            mock_get_notion_database.return_value = database_notion_data
            mock_get_database_notion_pages.return_value = database_pages_notion_data
            page = process_jinja_block(client, caption, jinja_code)
            pandoc_ast = page.to_pandoc()

    markdown = pandoc_ast_to_markdown(pandoc_ast)
    assert 'You attempted to access the "MISSING" property' in markdown
    assert 'the only available property is "title".' in markdown


@patch("n2y.notion.Client.wrap_notion_user")
def test_jinja_render_with_filter_error(wrap_notion_user):
    client = Client("")
    wrap_notion_user.return_value = User(client, mock_user())
    database_notion_data = mock_database(title="My DB")
    mention_notion_data = mock_database_mention(database_notion_data["id"])
    database_pages_notion_data = [mock_page(title="a"), mock_page(title="b")]
    caption = [
        mock_rich_text(spaced_format_def),
        mock_rich_text("My DB", mention=mention_notion_data),
    ]
    jinja_code = "{{ databases[0][0].title|round }}"

    with patch.object(Client, "get_notion_database") as mock_get_notion_database:
        with patch.object(
            Client, "get_database_notion_pages"
        ) as mock_get_database_notion_pages:
            mock_get_notion_database.return_value = database_notion_data
            mock_get_database_notion_pages.return_value = database_pages_notion_data
            page = process_jinja_block(client, caption, jinja_code)
            pandoc_ast = page.to_pandoc()

    markdown = pandoc_ast_to_markdown(pandoc_ast)
    assert 'Recieved the message "type str doesn\'t define __round__ method"' in markdown
    assert 'The Jinja filter "round" raised this error' in markdown
    assert 'argument(s): {\n\t"args": (\'a\',),\n\t"kwargs": {}\n}' in markdown


def test_jinja_render_plain():
    client = Client("")
    caption = mock_rich_text_array("{jinja=plain}")
    jinja_code = "# <h1> asdf23"
    page = process_jinja_block(client, caption, jinja_code)
    pandoc_ast = page.to_pandoc()
    text = pandoc_write_or_log_errors(pandoc_ast, "plain", [])
    assert text == "# <h1> asdf23\n"


def test_list_matches_special_characters():
    text = "somthing (id_num) [category] note* $tag (?!question)"
    string = (
        "fdsfds dwefwe 342t45lsfd fdds f34244 dgd somthing "
        "(id_num) [category] note* $tag (?!question) dsfsa "
        "sdfasdfsad"
    )
    assert len(list_matches(text, string)) == 1


def test_list_matches_no_space_fail():
    text = "somthing (id_num) [category] note* $tag (?!question)"
    string = (
        "fdsfds dwefwe 342t45lsfd fdds f34244 dgd sfdasomthing "
        "(id_num) [category] note* $tag (?!question) dsfsa "
        "sdfasdfsad"
    )
    assert len(list_matches(text, string)) == 0



import logging
import os
from os import listdir
from os.path import isfile, join

import yaml

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

from n2y.main import main
from n2y.notion import Client
from tests.utils import NOTION_ACCESS_TOKEN, parse_yaml_front_matter


def run_n2y(temp_dir, config):
    config_path = os.path.join(temp_dir, "config.yaml")
    with open(config_path, "w") as f:
        yaml.dump(config, f)
    old_cwd = os.getcwd()
    os.chdir(temp_dir)
    try:
        status = main([config_path], NOTION_ACCESS_TOKEN)
    finally:
        os.chdir(old_cwd)
    return status


def run_n2y_page(temp_dir, page_id, **export_config_keys):
    config = {
        "exports": [
            {
                "id": page_id,
                "node_type": "page",
                "output": "page.md",
                **export_config_keys,
            }
        ]
    }
    status = run_n2y(temp_dir, config)
    assert status == 0
    with open(str(temp_dir / "page.md"), "r") as f:
        page_as_markdown = f.read()
    return page_as_markdown


def run_n2y_custom(temp_dir, config):
    config_path = os.path.join(temp_dir, "config.yaml")
    with open(config_path, "w") as f:
        yaml.dump(config, f)
    old_cwd = os.getcwd()
    os.chdir(temp_dir)
    try:
        status = main([config_path], NOTION_ACCESS_TOKEN)
    finally:
        os.chdir(old_cwd)
    return status


def run_n2y_database_as_yaml(temp_dir, database_id, **export_config_keys):
    config = {
        "exports": [
            {
                "id": database_id,
                "node_type": "database_as_yaml",
                "output": "database.yml",
                **export_config_keys,
            }
        ]
    }
    status = run_n2y(temp_dir, config)
    assert status == 0
    with open(str(temp_dir / "database.yml"), "r") as f:
        unsorted_database = yaml.load(f, Loader=Loader)
    return unsorted_database


def run_n2y_database_as_files(temp_dir, database_id, **export_config_keys):
    config = {
        "exports": [
            {
                "id": database_id,
                "node_type": "database_as_files",
                "output": "database",
                **export_config_keys,
            }
        ]
    }
    status = run_n2y(temp_dir, config)
    assert status == 0
    return os.path.join(temp_dir, "database")


def test_simple_database_to_yaml(tmpdir):
    """
    The database can be seen here:
    https://fresh-pencil-9f3.notion.site/176fa24d4b7f4256877e60a1035b45a4
    """
    object_id = "176fa24d4b7f4256877e60a1035b45a4"
    unsorted_database = run_n2y_database_as_yaml(
        tmpdir, object_id, content_property="Content"
    )
    database = sorted(unsorted_database, key=lambda row: row["Name"])
    assert len(database) == 3
    assert database[0]["Name"] == "A"
    assert database[0]["Tags"] == ["a", "b"]
    assert database[0]["Content"] == "A\n"


def test_big_database_to_yaml(tmpdir):
    """
    The database can be seen here:
    https://fresh-pencil-9f3.notion.site/9341a0ddf7d4442d94ad74e5100f72af
    """
    object_id = "9341a0ddf7d4442d94ad74e5100f72af"
    database = run_n2y_database_as_yaml(tmpdir, object_id)
    assert len(database) == 101


def test_simple_database_to_markdown_files(tmpdir):
    """
    The database can be seen here:
    https://fresh-pencil-9f3.notion.site/176fa24d4b7f4256877e60a1035b45a4
    """
    object_id = "176fa24d4b7f4256877e60a1035b45a4"
    output_directory = run_n2y_database_as_files(
        tmpdir, object_id, filename_template="{Name}.md"
    )
    generated_files = {
        f for f in listdir(output_directory) if isfile(join(output_directory, f))
    }
    assert generated_files == {"A.md", "B.md", "C.md"}
    document = open(join(output_directory, "A.md"), "r").read()
    metadata = parse_yaml_front_matter(document)
    assert metadata["Name"] == "A"
    assert metadata["Tags"] == ["a", "b"]
    assert "content" not in metadata


def test_simple_database_to_docx_files(tmpdir):
    """
    The database can be seen here:
    https://fresh-pencil-9f3.notion.site/176fa24d4b7f4256877e60a1035b45a4
    """
    database_id = "176fa24d4b7f4256877e60a1035b45a4"
    config = {
        "exports": [
            {
                "id": database_id,
                "pandoc_format": "docx",
                "pandoc_options": ["--standalone"],
                "node_type": "database_as_files",
                "output": "database",
            }
        ]
    }
    status = run_n2y(tmpdir, config)
    assert status == 0
    output_directory = os.path.join(tmpdir, "database")
    generated_files = {
        f for f in listdir(output_directory) if isfile(join(output_directory, f))
    }
    assert generated_files == {"A.docx", "B.docx", "C.docx"}


def test_simple_database_config(tmpdir):
    """
    The database can be seen here:
    https://fresh-pencil-9f3.notion.site/176fa24d4b7f4256877e60a1035b45a4
    """
    database_id = "176fa24d4b7f4256877e60a1035b45a4"
    notion_sorts = [
        {
            "property": "Name",
            "direction": "descending",
        }
    ]
    notion_filter = {
        "or": [
            {"property": "Name", "rich_text": {"contains": "A"}},
            {"property": "Name", "rich_text": {"contains": "C"}},
        ]
    }
    database = run_n2y_database_as_yaml(
        tmpdir,
        database_id,
        notion_sort=notion_sorts,
        notion_filter=notion_filter,
    )
    assert len(database) == 2
    assert database[0]["Name"] == "C"
    assert database[1]["Name"] == "A"


def test_all_properties_database(tmpdir):
    """
    The page can be seen here:
    https://fresh-pencil-9f3.notion.site/53b9fa3da3f348e7ba3346254f1c722f
    """
    object_id = "53b9fa3da3f348e7ba3346254f1c722f"
    database = run_n2y_database_as_yaml(tmpdir, object_id)
    assert len(database) == 4


def test_mention_in_simple_table(tmpdir):
    """
    The page can be seen here:
    https://fresh-pencil-9f3.notion.site/Simple-Table-with-Mention-Test-e12497428b0e43c3b14e016de6c5a2cf
    """
    object_id = "e12497428b0e43c3b14e016de6c5a2cf"
    document = run_n2y_page(tmpdir, object_id)
    assert "In Table: Simple Test Page" in document
    assert "Out of Table: Simple Test Page" in document


def test_all_blocks_page_to_markdown(tmpdir):
    """
    The page can be seen here:
    https://fresh-pencil-9f3.notion.site/Test-Page-5f18c7d7eda44986ae7d938a12817cc0
    """
    object_id = "5f18c7d7eda44986ae7d938a12817cc0"
    document = run_n2y_page(tmpdir, object_id)
    lines = document.split("\n")
    metadata = parse_yaml_front_matter(document)
    assert metadata["title"] == "All Blocks Test Page"
    column_strings_in_lines = [
        "Column 1" in lines,
        "Column 1.1" in lines,
        "Column 1.2" in lines,
        "Column 2" in lines,
    ]
    assert lines[5:12] == [
        "1.  [Heading 1](#heading-1)",
        "    1.  [Heading 2](#heading-2)",
        "        1.  [Heading 3](#heading-3)",
        "2.  [Header One](#header-one)",
        "    1.  [Header Two](#header-two)",
        "        1.  [Header Three](#header-three)",
        "3.  [Columns](#columns)",
    ]
    assert "Text block" in lines
    assert "Text *italics* too" in lines
    assert "-   [ ] To do list block" in lines
    assert "# Heading 1" in lines
    assert "## Heading 2" in lines
    assert "### Heading 3" in lines
    assert "-   List block" in lines
    assert "1.  Number list block" in lines
    assert "-   Toggle list" in lines
    assert "> Block quote single paragraph" in lines
    assert "> Block quote second paragraph" in lines
    assert "---" in lines
    assert "Callout block" in lines
    assert "$e^{-i \\pi} = -1$" in lines
    assert "``` javascript\nCode Block\n```" in document
    assert lines.count("This is a synced block.") == 2
    assert "This is a synced block from another page." in lines
    assert all(column_strings_in_lines)
    assert "Mention: Simple Test Page" in lines
    assert "Simple Test Page" in lines  # from the LinkToPageBlock
    assert "Mention: Simple Test Database" in lines
    assert "Simple Test Database" in lines  # from the LinkToPageBlock

    # from the FileBlock
    assert any("[small_file.txt](media/All_Blocks_Test_Page-" in l for l in lines)

    # a bookmark with a caption and without
    assert "<https://innolitics.com>" in lines
    assert "[Bookmark caption](https://innolitics.com)" in lines

    # TODO: Add assertions about audio blocks and video blocks

    # the word "caption" is bolded
    assert "![Image **caption**](media/All_Blocks_Test_Page-5f1b0813453.jpeg)" in lines

    # from a file block in the Notion page
    assert os.path.exists(tmpdir / "media" / "All_Blocks_Test_Page-5f1b0813453.jpeg")


def test_page_in_database_to_markdown(tmpdir):
    """
    This test exports a single page, or "row", that is in a database. Unlike
    pages that are not in a database, who only have a single "Title" property,
    pages in a database will have properties for all of the "columns" in that
    database.

    The page can be seen here:
    https://fresh-pencil-9f3.notion.site/C-7e967a44893f4b25917965896e81c137
    """
    object_id = "7e967a44893f4b25917965896e81c137"
    document = run_n2y_page(tmpdir, object_id)
    lines = document.split("\n")
    metadata = parse_yaml_front_matter(document)
    assert metadata["Name"] == "C"
    assert metadata["Tags"] == ["d", "a", "b", "c"]
    assert "content" not in metadata
    assert "Has some basic content" in lines


def test_simple_page_to_markdown(tmpdir):
    """
    The page can be seen here:
    https://fresh-pencil-9f3.notion.site/Simple-Test-Page-6670dc17a7bc4426b91bca4cf3ac5623
    """
    object_id = "6670dc17a7bc4426b91bca4cf3ac5623"
    document = run_n2y_page(tmpdir, object_id)
    assert "Page content" in document


def test_builtin_plugins(tmpdir):
    """
    The page can be seen here:
    https://fresh-pencil-9f3.notion.site/Plugins-Test-96d71e2876eb47b285833582e8cf27eb
    """
    object_id = "96d71e2876eb47b285833582e8cf27eb"
    document = run_n2y_page(
        tmpdir,
        object_id,
        plugins=[
            "n2y.plugins.deepheaders",
            "n2y.plugins.removecallouts",
            "n2y.plugins.rawcodeblocks",
            "n2y.plugins.mermaid",
            "n2y.plugins.footnotes",
            "n2y.plugins.expandlinktopages",
        ],
    )
    lines = document.split("\n")

    assert "#### H4" in lines
    assert "##### H5" in lines
    # TODO: Fix failing mermaid assertions
    # assert not any("should disappear" in l for l in lines)
    # invalid_mermaid = '    invalid'
    # assert lines[15] != invalid_mermaid
    # if lines[15] == '    sequenceDiagram':
    #     assert lines[16] == '    A->>B: Hello'
    #     assert lines[18] == invalid_mermaid
    # else:
    #     assert lines[17] == invalid_mermaid
    #     assert 'media' in lines[15]
    #     im1_line = re.search(r'media.*png', lines[15])[0]
    #     im1 = f'{tmpdir.strpath}/{im1_line}'
    #     root = Path(__file__).resolve().parent.parent
    #     with open(im1, 'rb') as img:
    #         with open(root / 'n2y' / 'data' / 'mermaid_err.png', 'rb') as err:
    #             assert img.read() != err.read()

    assert "Raw markdown should show up" in lines
    assert "Raw html should not show up" not in lines

    # TODO: Fix these failing assertions
    # assert "# Header with Footnotes[^1]" in lines
    # assert "Paragraph with footnote.[^2]" in lines
    # assert "[^1]: First **footnote**." in lines
    # assert "[^2]: Second footnote" in lines

    # The word "Bulletlist" only shows up in the linked page that is expanded
    assert "Bulletlist" in document

    # Ensure a link to page to an unshared page doesn't get expanded; note that
    # Notion will actually represent these pages as an "UnsupportedBlock" (which
    # is odd). This will throw a warning and won't produce any content though,
    # which is the desired behavior.
    assert "Untitled" not in document
    assert "Unshared Page" not in document
    assert "This page is not shared with the integration." not in document


def test_comment():
    # TODO: Add a link to page where this block can be seen
    block_with_comments_id = "ac496c7db16743488495976f7433dbfb"
    comments = Client(NOTION_ACCESS_TOKEN).get_comments(block_with_comments_id)
    assert comments[0].rich_text.to_plain_text() == "Test Comment"
    assert comments[1].rich_text.to_plain_text() == "Test Comment 2"


def test_can_pull_all_relations(tmpdir):
    """
    This page can be seen here https://www.notion.so/Item-e611d7fdd4604d1192637a25bc9f5339

    The purpose of this test is to confirm that we're handling pagination
    unrolling for situations when a relation property has more than 30 items.
    """
    object_id = "e611d7fdd4604d1192637a25bc9f5339"
    page = run_n2y_page(tmpdir, object_id)
    page_properties = parse_yaml_front_matter(page)
    assert len(page_properties["Related Items"]) == 31


def test_jinja_render_plugin(tmpdir):
    """
    This test exercises our jinja render plugin. It involves a simple database
    with three entries and a jinja code block.

    See https://www.notion.so/Render-Test-4803ad13c6224deb93da94c41cc105d3 for
    the parent page with the database and child page to be rendered.
    """
    object_id = "869e37f1523c468399411a1b23b27634"
    document = run_n2y_page(
        tmpdir,
        object_id,
        plugins=[
            "n2y.plugins.jinjarenderpage",
        ],
    )

    assert "## blue is a color" in document
    assert "## pink is a color" in document
    assert "## red is a color" in document
    assert "Page With JinjaCodeBlock" in document


def test_dbfootnote_plugin(tmpdir):
    """
    This test exercises our database footnote. It involves a simple page with
    an inline database containing content for 3 footnotes.

    The page can be seen here:
    https://www.notion.so/Advanced-DB-Footnote-8a1a17c7ef3043f4bd9fb041ef31ba7
    """
    object_id = "8a1a17c7ef3043f4bd9fb041ef31ba7d"
    document = run_n2y_page(
        tmpdir,
        object_id,
        plugins=[
            "n2y.plugins.dbfootnotes",
        ],
    )

    assert "some text [^1]." in document
    assert "like so [^2] and" in document
    assert "so [^3]." in document
    assert "[^1]: Inline DB content for footnote 1." in document
    assert "[^2]: Inline DB content for footnote 2." in document
    assert "[^3]: Inline DB content for footnote 3." in document


def test_simple_table_headers(caplog, tmp_path):
    """
    Simply show what flattened Markdown content is expected if the input
    contains tables without headers.

    Relies on https://www.notion.so/Simple-Tables-9b1dd705f61647b6a10032ec7671402f?pvs=4
    """
    object_id = "9b1dd705f61647b6a10032ec7671402f"

    content = run_n2y_page(tmp_path, object_id, pandoc_format="gfm")
    captured_messages = [
        r.message for r in caplog.records if r.levelno >= logging.WARNING
    ]

    number_expected_table_warnings = 0
    if (
        """\
Some text

|      |         |
|------|---------|
| This | has     |
| no   | headers |
"""
        in content
    ):
        number_expected_table_warnings += 1
    assert (
        """\
More text

|        | header | row     |
|--------|--------|---------|
| header | This   | has     |
| column | both   | headers |
|        |        |         |
| and    | an     | empty   |
"""
        in content
    )
    if (
        """\
Yakkity yakkity yakkity yak

|        |        |
|--------|--------|
| header | Fiddle |
| column | Faddle |
"""
        in content
    ):
        number_expected_table_warnings += 1
    assert (
        """\
| header | row    |
|--------|--------|
| Nutter | Butter |
"""
        in content
    )
    if number_expected_table_warnings > 0:
        expected_blurb = (
            f"{number_expected_table_warnings} table(s) will present empty "
            "headers to maintain Markdown spec"
        )
        assert any(expected_blurb in m for m in captured_messages)



"""
Unit tests of the builtin plugins.
"""
import pytest

from n2y.blocks import ParagraphBlock, FencedCodeBlock
from n2y.notion import Client
from n2y.page import Page
from n2y.errors import PluginError, UseNextClass
from n2y.notion_mocks import mock_paragraph_block
from n2y.plugins.mermaid import MermaidFencedCodeBlock
from n2y.plugins.rawcodeblocks import RawFencedCodeBlock


def test_load_plugin_invalid_notion_object():
    client = Client("")
    with pytest.raises(PluginError) as err:
        client.load_plugin({"puppy": str})
    assert "puppy" in str(err)


def test_load_plugin_invalid_object_type():
    client = Client("")
    with pytest.raises(PluginError) as err:
        client.load_plugin({"blocks": {"puppy": str}})
    assert "block" in str(err)
    assert "puppy" in str(err)


def test_load_plugin_valid_page():
    client = Client("")

    class MyPage(Page):
        pass

    client.load_plugin({"page": MyPage})
    assert client.get_class_list("page") == [Page, MyPage]


def test_get_class_fallthrough():
    client = Client("")

    class SometimesParagraph(ParagraphBlock):
        def __init__(self, client, notion_data, page=None, get_children=False):
            super().__init__(client, notion_data, page, get_children)
            if self.rich_text.to_plain_text() != "sometimes":
                raise UseNextClass()

    client.load_plugin({"blocks": {"paragraph": SometimesParagraph}})
    sometimes = client.wrap_notion_block(
        mock_paragraph_block([("sometimes", {})]), None, False
    )
    othertimes = client.wrap_notion_block(
        mock_paragraph_block([("othertimes", {})]), None, False
    )
    assert type(sometimes) is SometimesParagraph
    assert type(othertimes) is ParagraphBlock


def test_load_plugin_invalid_page_class():
    client = Client("")

    class MyPage:
        pass

    with pytest.raises(PluginError) as err:
        client.load_plugin({"page": MyPage})
    assert "MyPage" in str(err)


def test_load_plugin_valid_block():
    client = Client("")

    class MyParagraphBlock(ParagraphBlock):
        pass

    client.load_plugin({"blocks": {"paragraph": MyParagraphBlock}})
    assert client.get_class_list("blocks", "paragraph") == [
        ParagraphBlock,
        MyParagraphBlock,
    ]


def test_load_plugin_invalid_block_mapping():
    client = Client("")
    with pytest.raises(PluginError) as err:
        client.load_plugin({"blocks": str})
    assert "block" in str(err)


def test_load_plugin_invalid_block_class():
    client = Client("")

    class MyParagraphBlock:
        pass

    with pytest.raises(PluginError) as err:
        client.load_plugin({"blocks": {"paragraph": MyParagraphBlock}})
    assert "MyParagraphBlock" in str(err)


def test_load_plugins_overrides_old_plugins():
    client = Client("")
    client.load_plugins(["n2y.plugins.mermaid"])
    assert client.get_class_list("blocks", "code") == [
        FencedCodeBlock,
        MermaidFencedCodeBlock,
    ]
    client.load_plugins(["n2y.plugins.rawcodeblocks"])
    assert client.get_class_list("blocks", "code") == [
        FencedCodeBlock,
        RawFencedCodeBlock,
    ]



import copy
import yaml

from n2y.config import (
    valid_notion_id,
    merge_config,
    load_config,
    _valid_notion_filter,
    _validate_config_item,
    EXPORT_DEFAULTS,
)
from n2y.notion_mocks import mock_id


def mock_config_item(node_type):
    config_item = copy.deepcopy(EXPORT_DEFAULTS)
    config_item["id"] = mock_id()
    config_item["node_type"] = node_type
    return config_item


def test_load_config_basic(tmp_path):
    # use a temporary file to test the config loading
    config_path = tmp_path / "config.yaml"
    export_id = mock_id()
    with open(config_path, "w") as f:
        f.write(
            yaml.dump(
                {
                    "media_root": "media",
                    "media_url": "https://example.com/media",
                    "export_defaults": {
                        "id_property": "id",
                        "url_property": "url",
                    },
                    "exports": [
                        {
                            "id": export_id,
                            "node_type": "page",
                            "output": "output.md",
                            "pandoc_format": "gfm",
                        }
                    ],
                }
            )
        )
    config = load_config(config_path)
    assert config is not None, "The config is invalid"
    merged_export = config["exports"][0]
    assert merged_export["id"] == export_id
    assert merged_export["node_type"] == "page"
    assert merged_export["id_property"] == "id"
    assert merged_export["url_property"] == "url"
    assert merged_export["pandoc_format"] == "gfm"


def test_merge_config_no_defaults():
    master_defaults = {"a": "1"}
    defaults = {}
    config_items = [
        {"b": "1"},
        {"b": "2", "a": "2"},
    ]
    assert merge_config(config_items, master_defaults, defaults) == [
        {"a": "1", "b": "1"},
        {"b": "2", "a": "2"},
    ]


def test_merge_config_defaults():
    master_defaults = {"a": "1", "b": "1"}
    defaults = {"a": "3"}
    config_items = [
        {},
        {"a": "2"},
        {"b": "2"},
    ]
    assert merge_config(config_items, master_defaults, defaults) == [
        {"a": "3", "b": "1"},
        {"a": "2", "b": "1"},
        {"a": "3", "b": "2"},
    ]


def test_valid_id_valid():
    assert valid_notion_id(mock_id())


def test_valid_id_invalid():
    assert not valid_notion_id(mock_id() + "a")


def test_valid_id_invalid_due_to_special():
    bad_id = "https://" + mock_id()[8:]
    assert not valid_notion_id(bad_id)


def test_valid_notion_filter_simple():
    assert _valid_notion_filter(
        {
            "property": "title",
            "direction": "ascending",
        }
    )


def test_valid_notion_filter_complex():
    assert _valid_notion_filter(
        [
            {
                "property": "title",
                "direction": "ascending",
            }
        ]
    )


def test_valid_config_item_missing_id():
    config_item = mock_config_item("page")
    del config_item["id"]
    assert not _validate_config_item(config_item)


def test_valid_config_item_missing_node_type():
    config_item = mock_config_item("page")
    del config_item["node_type"]
    assert not _validate_config_item(config_item)


def test_valid_config_item_invalid_node_type():
    config_item = mock_config_item("page")
    config_item["node_type"] = "invalid"
    assert not _validate_config_item(config_item)


def test_valid_config_item_missing_filename_template():
    config_item = mock_config_item("database_as_files")
    assert not _validate_config_item(config_item)


def test_valid_config_item_malformed_filename_template():
    config_item = mock_config_item("database_as_files")
    config_item["filename_template"] = "{"
    assert not _validate_config_item(config_item)



import sys
from io import StringIO

from tests.utils import NOTION_ACCESS_TOKEN
from n2y.audit import main


def run_n2yaudit(arguments):
    old_stdout = sys.stdout
    sys.stdout = StringIO()
    try:
        status = main(arguments, NOTION_ACCESS_TOKEN)
        stdout = sys.stdout.getvalue()
    finally:
        sys.stdout = old_stdout
    return status, stdout


def test_audit():
    """
    The database can be seen here:
    https://fresh-pencil-9f3.notion.site/Audited-cfa8ff07bba244c8b967c9b6a7a954c1
    """
    object_id = "cfa8ff07bba244c8b967c9b6a7a954c1"
    status, stdoutput = run_n2yaudit([object_id])
    assert status == 3

    external_mention_in_top_page = "https://www.notion.so/Audited-cfa8ff07bba244c8b967c9b6a7a954c1#aa4fa886f8244c818de8018bb3491806"  # noqa: E501
    external_mention_in_child_page = "https://www.notion.so/Child-f3e3628fc80c470ea68994fa7ec0ff17#d1d32ff6f0cb4c71a2f1c4ec55e00086"  # noqa: E501
    internal_mention_in_child_page = "https://www.notion.so/Child-f3e3628fc80c470ea68994fa7ec0ff17#eab91ccc32924221ac3f0a74225a33dd"  # noqa: E501
    external_mention_in_child_database = "https://www.notion.so/B-4412005dcec24ff2827abbc367c90b29#6373a0b5c2804fbe9dfac167ce6948a0"  # noqa: E501
    internal_mention_in_database_in_column = "https://www.notion.so/Audited-cfa8ff07bba244c8b967c9b6a7a954c1#21a13c06ef86462e882a181c6cb52a64"  # noqa: E501

    # NOTE: When you try to get a link for a "LinkToPageBlock" in the Notion UI,
    # it appears to give you the URL for the linked page or database, and not to
    # the block itself. Thus, these two links were extracted using a debugger
    # when running this test.

    external_link_in_top_page = "https://www.notion.so/Audited-cfa8ff07bba244c8b967c9b6a7a954c1#c22d76d50c704761b0e729531e6cc24b"  # noqa: E501
    internal_link_in_top_page = "https://www.notion.so/Audited-cfa8ff07bba244c8b967c9b6a7a954c1#3e2e2fb2e9bf4a1f8b00bbb18a9d97e9"  # noqa: E501

    assert external_mention_in_top_page in stdoutput
    assert external_mention_in_child_page in stdoutput
    assert internal_mention_in_child_page not in stdoutput
    assert external_mention_in_child_database in stdoutput
    assert internal_mention_in_database_in_column not in stdoutput
    assert external_link_in_top_page in stdoutput
    assert internal_link_in_top_page not in stdoutput



from pandoc.types import (
    Str,
    Space,
    Strong,
    Emph,
    Strikeout,
    Code,
    Link,
    Math,
    InlineMath,
    Underline,
)

from n2y.notion import Client
from n2y.rich_text import RichTextArray, MentionRichText
from n2y.mentions import PageMention
from n2y.notion_mocks import (
    mock_rich_text_array,
    mock_annotations,
    mock_rich_text,
    mock_id,
)

word = "word."
spaced_word = " word."
bold_word = "A bold word."


def process_rich_text_array(notion_data):
    client = Client("")
    rich_text_array = client.wrap_notion_rich_text_array(notion_data)
    pandoc_ast = rich_text_array.to_pandoc()
    markdown = rich_text_array.to_value("markdown")
    plain_text = rich_text_array.to_plain_text()
    return pandoc_ast, markdown, plain_text


def test_bold_word():
    notion_data = mock_rich_text_array(
        [("A ", []), ("bold", ["bold"]), (spaced_word, [])]
    )
    pandoc_ast, markdown, plain_text = process_rich_text_array(notion_data)
    assert pandoc_ast == [Str("A"), Space(), Strong([Str("bold")]), Space(), Str(word)]
    assert markdown == "A **bold** word.\n"
    assert plain_text == bold_word


def test_bold_letter():
    notion_data = mock_rich_text_array([("A ", []), ("b", ["bold"]), ("old word.", [])])
    pandoc_ast, markdown, plain_text = process_rich_text_array(notion_data)
    assert pandoc_ast == [
        Str("A"),
        Space(),
        Strong([Str("b")]),
        Str("old"),
        Space(),
        Str(word),
    ]
    assert markdown == "A **b**old word.\n"
    assert plain_text == bold_word


def test_bold_spaces():
    notion_data = mock_rich_text_array([("A", []), (" bold ", ["bold"]), (word, [])])
    pandoc_ast, markdown, plain_text = process_rich_text_array(notion_data)
    assert pandoc_ast == [Str("A"), Space(), Strong([Str("bold")]), Space(), Str(word)]
    assert markdown == "A **bold** word.\n"
    assert plain_text == bold_word


def test_italic_word():
    notion_data = mock_rich_text_array(
        [("An ", []), ("italic", ["italic"]), (spaced_word, [])]
    )
    pandoc_ast, markdown, _ = process_rich_text_array(notion_data)
    assert pandoc_ast == [Str("An"), Space(), Emph([Str("italic")]), Space(), Str(word)]
    assert markdown == "An *italic* word.\n"


def test_italic_letter():
    notion_data = mock_rich_text_array(
        [("An ", []), ("i", ["italic"]), ("talic word.", [])]
    )
    pandoc_ast, markdown, _ = process_rich_text_array(notion_data)
    assert pandoc_ast == [
        Str("An"),
        Space(),
        Emph([Str("i")]),
        Str("talic"),
        Space(),
        Str(word),
    ]
    assert markdown == "An *i*talic word.\n"


def test_bold_italic_word():
    notion_data = mock_rich_text_array(
        [
            ("A ", []),
            ("bold-italic", ["bold", "italic"]),
            (spaced_word, []),
        ]
    )
    pandoc_ast, markdown, plain_text = process_rich_text_array(notion_data)
    assert pandoc_ast == [
        Str("A"),
        Space(),
        Emph([Strong([Str("bold-italic")])]),
        Space(),
        Str(word),
    ]
    assert markdown == "A ***bold-italic*** word.\n"
    assert plain_text == "A bold-italic word."


def test_bold_space():
    notion_data = mock_rich_text_array(
        [
            (" ", ["bold"]),
        ]
    )
    pandoc_ast, _, _ = process_rich_text_array(notion_data)
    assert pandoc_ast == [Space()]


def test_italic_spaces():
    notion_data = mock_rich_text_array(
        [("An", []), (" italic ", ["italic"]), (word, [])]
    )
    pandoc_ast, markdown, plain_text = process_rich_text_array(notion_data)
    assert pandoc_ast == [Str("An"), Space(), Emph([Str("italic")]), Space(), Str(word)]
    assert markdown == "An *italic* word.\n"
    assert plain_text == "An italic word."


def test_strikeout_word():
    notion_data = mock_rich_text_array(
        [("A ", []), ("deleted", ["strikethrough"]), (spaced_word, [])]
    )
    pandoc_ast, markdown, plain_text = process_rich_text_array(notion_data)
    assert pandoc_ast == [
        Str("A"),
        Space(),
        Strikeout([Str("deleted")]),
        Space(),
        Str(word),
    ]
    assert markdown == "A ~~deleted~~ word.\n"
    assert plain_text == "A deleted word."


def test_strikeout_spaces():
    notion_data = mock_rich_text_array(
        [("A", []), (" deleted ", ["strikethrough"]), (word, [])]
    )
    pandoc_ast, markdown, _ = process_rich_text_array(notion_data)
    assert pandoc_ast == [
        Str("A"),
        Space(),
        Strikeout([Str("deleted")]),
        Space(),
        Str(word),
    ]
    assert markdown == "A ~~deleted~~ word.\n"


def test_blended_annotated_spaces():
    notion_data = mock_rich_text_array(
        [
            ("this ", ["bold"]),
            ("is ", ["bold", "italic"]),
            ("a", ["italic"]),
            (" test", ["strikethrough"]),
            (" did", ["bold", "underline", "code"]),
            (" i pass", ["underline", "code"]),
            ("?", []),
        ]
    )
    pandoc_ast, markdown, _ = process_rich_text_array(notion_data)
    assert pandoc_ast == [
        Strong([Str("this")]),
        Space(),
        Emph([Strong([Str("is")])]),
        Space(),
        Emph([Str("a")]),
        Space(),
        Strikeout([Str("test")]),
        Underline([Strong([Code(("", [], []), " did")])]),
        Underline([Code(("", [], []), " i pass")]),
        Str("?"),
    ]
    assert (
        markdown == "**this** ***is*** *a*\n~~test~~[**` did`*"
        "*]{.underline}[` i pass`]{.underline}?\n"
    )


def test_equation_inline():
    lhs = "{\\displaystyle i\\hbar {\\frac {d}{dt}}\\vert "
    rhs = "\\Psi (t)\\rangle={\\hat {H}}\\vert \\Psi (t)\\rangle}"
    example_equation = f"{lhs}{rhs}"

    notion_data = [
        mock_rich_text("The Schrödinger Equation ("),
        {
            "type": "equation",
            "equation": {"expression": example_equation},
            "annotations": mock_annotations(),
            "plain_text": example_equation,
            "href": None,
        },
        mock_rich_text(") is a very useful one indeed"),
    ]

    pandoc_ast, markdown, _ = process_rich_text_array(notion_data)
    assert pandoc_ast == [
        Str("The"),
        Space(),
        Str("Schrödinger"),
        Space(),
        Str("Equation"),
        Space(),
        Str("("),
        Math(InlineMath(), example_equation),
        Str(")"),
        Space(),
        Str("is"),
        Space(),
        Str("a"),
        Space(),
        Str("very"),
        Space(),
        Str("useful"),
        Space(),
        Str("one"),
        Space(),
        Str("indeed"),
    ]
    assert (
        markdown == "The Schrödinger Equation\n(${\\displaystyle "
        "i\\hbar {\\frac {d}{dt}}\\vert \\Psi (t)\\"
        "rangle={\\hat {H}}\\vert \\Psi (t)\\rangle}$)\nis"
        " a very useful one indeed\n"
    )


def test_code_inline():
    notion_data = mock_rich_text_array(
        [("A ", []), ("code", ["code"]), (spaced_word, [])]
    )
    pandoc_ast, markdown, plain_text = process_rich_text_array(notion_data)
    assert pandoc_ast == [
        Str("A"),
        Space(),
        Code(("", [], []), "code"),
        Space(),
        Str(word),
    ]
    assert markdown == "A `code` word.\n"
    assert plain_text == "A code word."


def test_link_inline():
    notion_data = [
        mock_rich_text("This is a "),
        mock_rich_text("link", ["bold"], href="https://example.com"),
        mock_rich_text("."),
    ]
    pandoc_ast, markdown, plain_text = process_rich_text_array(notion_data)
    assert pandoc_ast == [
        Str("This"),
        Space(),
        Str("is"),
        Space(),
        Str("a"),
        Space(),
        Link(("", [], []), [Strong([Str("link")])], ("https://example.com", "")),
        Str("."),
    ]
    assert markdown == "This is a [**link**](https://example.com).\n"
    assert plain_text == "This is a link."


def test_prepend():
    client = Client("")
    plain_data = mock_rich_text_array("plain text")
    mention_data = {"type": "page", "page": {"id": mock_id()}}
    mention_text = "Test Page"
    mention_block = PageMention(client, mention_data, mention_text)
    mention_rich_text = MentionRichText(
        client,
        mock_rich_text(mention_text, None, None, mention_data),
        None,
        mention_block,
    )
    equation_text = "e=mc^2"
    equation_data = mock_rich_text_array(
        [[equation_text, None, None, None, None, {"expression": equation_text}]]
    )
    plain = RichTextArray(client, plain_data)
    mention = RichTextArray(client, [])
    equation = RichTextArray(client, equation_data)
    mention.items.append(mention_rich_text)
    assert plain.to_pandoc() == [Str("plain"), Space(), Str("text")]
    assert mention.to_pandoc() == [Str("Test"), Space(), Str("Page")]
    assert equation.to_pandoc() == [Math(InlineMath(), equation_text)]



class User:
    def __init__(self, client, notion_data):
        self.client = client
        self.notion_data = notion_data
        self.notion_id = notion_data["id"]
        self.notion_type = notion_data.get("type", None)
        self.name = notion_data.get("name", None)
        self.avatar_url = notion_data.get("avatar_url", None)
        if self.notion_type == "person" and "email" in notion_data["person"]:
            self.email = notion_data["person"]["email"]
        else:
            self.email = None
        # TODO: handle bot users

    def to_value(self):
        # TODO: consider how to handle people with the same name. E.g., maybe we
        # should at least show a warning
        return self.name or ""



from math import isclose
from pytest import raises
from datetime import datetime, timezone, timedelta

import pytest
from pandoc.types import MetaMap, MetaList, MetaBool, MetaString

from n2y.errors import APIResponseError
from n2y.utils import (
    fromisoformat,
    header_id_from_text,
    id_from_share_link,
    retry_api_call,
    yaml_to_meta_value,
)


def test_fromisoformat_datetime():
    expected = datetime(2022, 5, 10, 19, 52, tzinfo=timezone.utc)
    assert fromisoformat("2022-05-10T19:52:00.000Z") == expected


def test_fromisoformat_date():
    expected = datetime(2022, 5, 11)
    assert fromisoformat("2022-05-11") == expected


def test_database_id_from_share_link_id():
    database_id = "f77db3af4ab6a47b6162dacd76681231"
    assert id_from_share_link(database_id) == database_id


def test_database_id_from_share_link_id_hyphens():
    database_id = "f77db3af4ab6a47b6162dacd76681231"
    database_id_hyphens = "f77db3af-4ab6a47b6162dacd-76681231"
    assert id_from_share_link(database_id_hyphens) == database_id


def test_database_id_from_share_link_no_hyphens():
    database_id = "f77db3af4ab6a47b6162dacd76681231"
    view_id = "463a296148dc38f791e7037dda9a8c3f"
    share_link = f"https://www.notion.so/ws/{database_id}?v={view_id}"
    assert id_from_share_link(share_link) == database_id


def test_database_id_from_share_link_hyphens():
    database_id = "90a3f77db3af4ab6a47b6162dacd1111"
    database_id_hyphens = "90a3f77-db3af4ab6a47b6162-dacd1111"
    view_id = "463a296148dc38f791e7037dda9a8c3f"
    share_link = f"https://www.notion.so/ws/{database_id_hyphens}?v={view_id}"
    assert id_from_share_link(share_link) == database_id


def test_page_id_from_share_link():
    page_id = "42361d2334624e949d9b7e94d30c988b"
    share_link = f"https://www.notion.so/innolitics/SaMD-DHF-Template-{page_id}"
    assert id_from_share_link(share_link) == page_id


class MockResponse:
    def __init__(self, time, code):
        self.headers = {"retry-after": time}
        self.text = ""
        self.status_code = code


def test_retry_api_call_no_error():
    @retry_api_call
    def tester():
        return True

    assert tester()


def test_retry_api_call_multiple_errors():
    status_code = 429

    call_count = 0

    @retry_api_call
    def tester(time):
        seconds = timedelta.total_seconds(datetime.now() - time)
        nonlocal call_count
        call_count += 1

        if call_count == 1:
            raise APIResponseError(MockResponse(0.05, status_code), "", status_code)
        elif call_count == 2:
            assert isclose(0.05, seconds, abs_tol=0.1)
            raise APIResponseError(MockResponse(0.23, status_code), "", status_code)
        elif call_count == 3:
            assert isclose(0.35, seconds, abs_tol=0.1)
            raise APIResponseError(MockResponse(0.16, status_code), "", status_code)
        elif call_count == 4:
            assert isclose(0.51, seconds, abs_tol=0.1)
            return True

    assert tester(datetime.now())


@pytest.mark.parametrize("status_code", [409, 429, 500, 502, 504])
def test_retry_api_call_onc(status_code):
    call_count = 0

    @retry_api_call
    def tester():
        nonlocal call_count
        call_count += 1

        if call_count == 1:
            raise APIResponseError(MockResponse(0.001, status_code), "", status_code)
        else:
            return True

    assert tester()
    assert call_count == 2


def test_retry_api_call_max_errors():
    status_code = 429

    @retry_api_call
    def tester():
        raise APIResponseError(MockResponse(0.001, status_code), "", status_code)

    with raises(APIResponseError):
        tester()


def test_yaml_to_meta_value_scalar():
    assert yaml_to_meta_value("test") == MetaString("test")
    assert yaml_to_meta_value(3) == MetaString("3")
    assert yaml_to_meta_value(3.45) == MetaString("3.45")
    assert yaml_to_meta_value(True) == MetaBool(True)
    assert yaml_to_meta_value(False) == MetaBool(False)
    assert yaml_to_meta_value(None) == MetaString("")


def test_yaml_to_meta_value_map():
    assert yaml_to_meta_value({"a": "1", "b": 2}) == MetaMap(
        {
            "a": MetaString("1"),
            "b": MetaString("2"),
        }
    )


def test_yaml_to_meta_value_list():
    assert yaml_to_meta_value(["a", 1]) == MetaList([MetaString("a"), MetaString("1")])


def test_header_id_from_text_basic():
    assert header_id_from_text("Essays") == "essays"
    assert header_id_from_text("Hello Goodbye") == "hello-goodbye"


def test_header_id_from_text():
    """
    Most of these test cases were taken from the pandoc documentation. See:
    https://pandoc.org/MANUAL.html#extension-auto_identifiers
    """
    assert (
        header_id_from_text("Heading identifiers in HTML")
        == "heading-identifiers-in-html"
    )
    assert header_id_from_text("Maître d'hôtel") == "maître-dhôtel"
    assert header_id_from_text("*Dogs*?--in *my* house?") == "dogs--in-my-house"
    assert header_id_from_text("[HTML], [S5], or [RTF]?") == "html-s5-or-rtf"
    assert header_id_from_text("3. Applications") == "applications"
    assert header_id_from_text("33") == "section"

    # These are additional tests constructed by us.
    assert header_id_from_text("33 section") == "section"
    assert header_id_from_text("") == "section"
    assert header_id_from_text("newlines\nare hyphens") == "newlines-are-hyphens"


def test_header_id_from_text_existing_ids():
    assert header_id_from_text("a", {"a"}) == "a-1"
    assert header_id_from_text("a", {"a", "a-1"}) == "a-2"
    assert header_id_from_text("", {"section", "section-1"}) == "section-2"
    assert header_id_from_text("", {"section", "a", "a-1"}) == "section-1"



from n2y.logger import logger


class File:
    """
    See https://developers.notion.com/reference/file-object
    """

    def __init__(self, client, notion_data):
        self.client = client
        if notion_data["type"] == "file":
            logger.debug('Instantiating file "%s"', notion_data["file"]["url"])
            self.type = "file"
            self.url = notion_data["file"]["url"]
            self.expiry_time = notion_data["file"]["expiry_time"]
        elif notion_data["type"] == "external":
            logger.debug(
                'Instantiating external file "%s"', notion_data["external"]["url"]
            )
            self.type = "external"
            self.url = notion_data["external"]["url"]
        else:
            file_type = notion_data["type"]
            raise ValueError(f'Unknown file type "{file_type}"')

    def to_value(self):
        return self.url



from pandoc.types import Note

from n2y.database import Database
from n2y.errors import PluginError, UseNextClass
from n2y.mentions import PageMention

plugin_data_key = "n2y.plugins.dbfootnotes"


class PageMentionFootnote(PageMention):
    def __init__(self, client, notion_data, plain_text, block):
        super().__init__(client, notion_data, plain_text, block)
        if block is not None and self._references_correct() and self._is_footnote():
            self._attach_footnote()
        else:
            raise UseNextClass

    def _get_mention(self):
        # This should be the footnote content page.
        self.mentioned_page = self.client.get_page(self.notion_page_id)
        return self.mentioned_page is not None

    def _get_parents(self):
        # This should be the inline database.
        self.mentioned_page_parent = self.mentioned_page.parent
        # This should be the same page that houses the original mention.
        self.mentioned_page_parent_parent = self.mentioned_page_parent.parent
        return (
            self.mentioned_page_parent is not None
            and self.mentioned_page_parent_parent is not None
        )

    def _references_correct(self):
        return self._get_mention() and self._get_parents()

    def _is_footnote(self):
        # Check that the footnote parent is a DB with the "Footnotes" suffix in the title.
        if isinstance(
            self.mentioned_page_parent, Database
        ) and self.mentioned_page_parent.title.to_plain_text().endswith("Footnotes"):
            # Raise an exception if the footnotes DB is not a child of the original page.
            # I.e., it's a valid Footnotes DB but not a child (inline) with the original page.
            # This could occur if the user referenced a Footnotes DB on a different page, which
            # we explicitly do not support.
            if (
                self.mentioned_page_parent_parent.notion_data["id"]
                != self.block.page.notion_data["id"]
            ):
                try:
                    footnote_identifier = (
                        f"title: {self.mentioned_page.title.to_plain_text()}"
                    )
                except AttributeError:
                    footnote_identifier = f"id: {self.mentioned_page.notion_data['id']}"

                raise PluginError(
                    f"Using plugin `{plugin_data_key}`: "
                    f"For footnote ({footnote_identifier}) mention "
                    f"on page: {self.block.page.notion_data['url']}, "
                    "the associated footnotes DB "
                    f"({self.mentioned_page_parent.notion_data['url']}) "
                    "is not a child of the original page"
                )
            return True
        else:
            return False

    def _attach_footnote(self):
        block_content = self.mentioned_page.block.children_to_pandoc()
        self._footnote_in_pandoc_ast = (
            Note(block_content)
            if isinstance(block_content, list)
            else Note([block_content])
        )

    def to_pandoc(self):
        return [self._footnote_in_pandoc_ast]


notion_classes = {"mentions": {"page": PageMentionFootnote}}



class Property:
    def __init__(self, client, notion_data):
        self.client = client
        self.notion_id = notion_data["id"]
        self.notion_type = notion_data["type"]
        self.name = notion_data["name"]


# TODO: eventually add methods for generating SQL type expresions or JSON
# schema from these guys


class TitleProperty(Property):
    pass


class TextProperty(Property):
    pass


class NumberProperty(Property):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        self.format = notion_data["number"]["format"]


class StatusProperty(Property):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        notion_groups = notion_data["status"]["groups"]
        notion_options = notion_data["status"]["options"]
        self.groups = [StatusGroup(self.client, ng) for ng in notion_groups]
        self.options = [StatusOption(self.client, no) for no in notion_options]


class StatusOption:
    def __init__(self, client, notion_option):
        self.client = client
        self.notion_id = notion_option["id"]
        self.name = notion_option["name"]
        self.color = notion_option["color"]


class StatusGroup:
    def __init__(self, client, notion_group):
        self.client = client
        self.name = notion_group["name"]
        self.color = notion_group["color"]
        self.notion_id = notion_group["id"]
        self.option_ids = notion_group["option_ids"]


class SelectProperty(Property):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        notion_options = notion_data["select"]["options"]
        self.options = [SelectOption(self.client, no) for no in notion_options]


class SelectOption:
    def __init__(self, client, notion_option):
        self.client = client
        self.notion_id = notion_option["id"]
        self.name = notion_option["name"]
        self.color = notion_option["color"]


class MultiSelectProperty(Property):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        notion_options = notion_data["multi_select"]["options"]
        self.options = [MultiSelectOption(self.client, no) for no in notion_options]


class MultiSelectOption:
    """
    This class appears identical to SelectOption, but the notion API docs calls it a
    separate thing so it appears like it may change in the future; also, it
    seems like one may want a plugin that treats the two types of options
    differently
    """

    def __init__(self, client, notion_option):
        self.client = client
        self.notion_id = notion_option["id"]
        self.name = notion_option["name"]
        self.color = notion_option["color"]


class DateProperty(Property):
    pass


class PeopleProperty(Property):
    pass


class FilesProperty(Property):
    pass


class CheckboxProperty(Property):
    pass


class UrlProperty(Property):
    pass


class EmailProperty(Property):
    pass


class PhoneNumberProperty(Property):
    pass


class FormulaProperty(Property):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        self.formula = notion_data["formula"]["expression"]


class RelationProperty(Property):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        relation = notion_data["relation"]
        self.database_id = relation["database_id"]
        self.synced_property_name = relation.get("synced_property_name", None)
        self.synced_property_id = relation.get("synced_property_id", None)


class RollupProperty(Property):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        rollup = notion_data["rollup"]
        self.relation_property_name = rollup["relation_property_name"]
        self.relation_property_id = rollup["relation_property_id"]
        self.rollup_property_name = rollup["rollup_property_name"]
        self.rollup_property_id = rollup["rollup_property_id"]
        self.function = rollup["function"]


class CreatedTimeProperty(Property):
    pass


class CreatedByProperty(Property):
    pass


class LastEditedTimeProperty(Property):
    pass


class LastEditedBy(Property):
    pass


DEFAULT_PROPERTIES = {
    "title": TitleProperty,
    "rich_text": TextProperty,
    "number": NumberProperty,
    "status": StatusProperty,
    "select": SelectProperty,
    "multi_select": MultiSelectProperty,
    "date": DateProperty,
    "people": PeopleProperty,
    "files": FilesProperty,
    "checkbox": CheckboxProperty,
    "url": UrlProperty,
    "email": EmailProperty,
    "phone_number": PhoneNumberProperty,
    "formula": FormulaProperty,
    "relation": RelationProperty,
    "rollup": RollupProperty,
    "created_time": CreatedTimeProperty,
    "created_by": CreatedByProperty,
    "last_edited_time": LastEditedTimeProperty,
    "last_edited_by": LastEditedBy,
}



from n2y.blocks import ToggleBlock
from n2y.errors import UseNextClass


plugin_data_key = "n2y.plugins.hiddenjinjatoggles"


class ExpandBlueToggleBlock(ToggleBlock):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        if self.notion_type_data["color"] != "blue_background":
            raise UseNextClass()

    def to_pandoc(self):
        return self.children_to_pandoc()


notion_classes = {
    "blocks": {"toggle": ExpandBlueToggleBlock},
}



import re

from pandoc.types import RawBlock, Format

from n2y.blocks import FencedCodeBlock
from n2y.errors import UseNextClass


class RawFencedCodeBlock(FencedCodeBlock):
    """
    Adds support for raw codeblocks.

    Any code block whose caption begins with "{=language}" will be made into a
    raw block for pandoc to parse. This is useful if you need to drop into Raw
    HTML or other formats.

    See https://pandoc.org/MANUAL.html#generic-raw-attribute
    """

    trigger_regex = re.compile(r"^{=(.+)}")

    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        result = self.caption.matches(self.trigger_regex)
        if result:
            self.raw_lang = result.group(1)
        else:
            raise UseNextClass()

    def to_pandoc(self):
        return RawBlock(Format(self.raw_lang), self.rich_text.to_plain_text())


notion_classes = {
    "blocks": {
        "code": RawFencedCodeBlock,
    }
}



"""
These tests verify how the n2y block classes convert notion data into Pandoc
abstract syntax tree (AST) objects, and then into markdown.
"""
import re
from unittest import mock

import pytest
from pandoc.types import (
    AlignDefault,
    BlockQuote,
    BulletList,
    Caption,
    Cell,
    CodeBlock,
    ColSpan,
    ColWidthDefault,
    Decimal,
    DisplayMath,
    Header,
    HorizontalRule,
    Image,
    LineBreak,
    Link,
    Math,
    Meta,
    OrderedList,
    Pandoc,
    Para,
    Period,
    Plain,
    Row,
    RowHeadColumns,
    RowSpan,
    Space,
    Str,
    Table,
    TableBody,
    TableFoot,
    TableHead,
)

from n2y.notion import Client
from n2y.notion_mocks import (
    mock_block,
    mock_file,
    mock_id,
    mock_page,
    mock_page_mention,
    mock_paragraph_block,
    mock_rich_text,
    mock_rich_text_array,
    mock_user,
)
from n2y.page import Page
from n2y.user import User
from n2y.utils import pandoc_ast_to_markdown

innolitics_website = "https://innolitics.com"
example_img = "https://example.com/image.png"
toc_headers = [
    Header(
        1,
        ("foo-items-header", [], []),
        [Str("Foo"), Space(), Str("Items"), Space(), Str("Header")],
    ),
    Header(2, ("foo-bar", [], []), [Str("Foo:"), Space(), Str("Bar")]),
    Header(2, ("foo-ski", [], []), [Str("Foo:"), Space(), Str("Ski")]),
    Header(
        3,
        ("first-foo-ski", [], []),
        [Str("First"), Space(), Str("Foo"), Space(), Str("Ski")],
    ),
]
toc_headers_starting_h2 = [
    Header(
        2,
        ("foo-header-h2", [], []),
        [Str("Foo"), Space(), Str("Header"), Space(), Str("H2")],
    ),
    Header(
        3,
        ("foo-bar-h3", [], []),
        [Str("Foo:"), Space(), Str("Bar"), Space(), Str("H3")],
    ),
    Header(
        3,
        ("foo-ski-h3", [], []),
        [Str("Foo:"), Space(), Str("Ski"), Space(), Str("H3")],
    ),
    Header(
        4,
        ("first-foo-ski", [], []),
        [Str("First"), Space(), Str("Foo"), Space(), Str("Ski")],
    ),
]
toc_item_ast = [
    Plain(
        [
            Link(
                ("", [], []),
                [Str("Foo"), Space(), Str("Items"), Space(), Str("Header")],
                ("#foo-items-header", ""),
            )
        ]
    ),
    OrderedList(
        (1, Decimal(), Period()),
        [
            [
                Plain(
                    [
                        Link(
                            ("", [], []),
                            [Str("Foo:"), Space(), Str("Bar")],
                            ("#foo-bar", ""),
                        )
                    ]
                )
            ],
            [
                Plain(
                    [
                        Link(
                            ("", [], []),
                            [Str("Foo:"), Space(), Str("Ski")],
                            ("#foo-ski", ""),
                        )
                    ]
                ),
                OrderedList(
                    (1, Decimal(), Period()),
                    [
                        [
                            Plain(
                                [
                                    Link(
                                        ("", [], []),
                                        [
                                            Str("First"),
                                            Space(),
                                            Str("Foo"),
                                            Space(),
                                            Str("Ski"),
                                        ],
                                        ("#first-foo-ski", ""),
                                    )
                                ]
                            )
                        ]
                    ],
                ),
            ],
        ],
    ),
]


def generate_block(notion_block, plugins=None):
    with mock.patch.object(Client, "get_notion_block") as mock_get_notion_block:
        mock_get_notion_block.return_value = notion_block
        client = Client("", plugins=plugins)
        page = None
        return client.get_block("unusedid", page)


def process_block(notion_block, plugins=None):
    n2y_block = generate_block(notion_block, plugins=plugins)
    pandoc_ast = n2y_block.to_pandoc()
    markdown = pandoc_ast_to_markdown(pandoc_ast)
    return pandoc_ast, markdown


def process_parent_block(notion_block, child_notion_blocks, plugins=None):
    with mock.patch.object(
        Client, "get_child_notion_blocks"
    ) as mock_get_child_notion_blocks:
        mock_get_child_notion_blocks.return_value = child_notion_blocks
        n2y_block = generate_block(notion_block, plugins)
    pandoc_ast = n2y_block.to_pandoc()
    markdown = pandoc_ast_to_markdown(pandoc_ast)
    return pandoc_ast, markdown


def test_unknown_block_type():
    notion_block = mock_block("abcdef", {})
    with pytest.raises(NotImplementedError) as e:
        process_block(notion_block)
    assert "abcdef" in str(e)


def test_paragraph():
    notion_block = mock_paragraph_block([("paragraph text", [])])
    pandoc_ast, markdown = process_block(notion_block)
    assert pandoc_ast == Para([Str("paragraph"), Space(), Str("text")])
    assert markdown == "paragraph text\n"


def test_paragraph_children_have_block_references():
    notion_block = mock_block(
        "paragraph",
        {
            "color": "default",
            "rich_text": [mock_rich_text("m", mention=mock_page_mention())],
        },
    )
    n2y_block = generate_block(notion_block)
    assert n2y_block.rich_text.block == n2y_block
    assert n2y_block.rich_text[0].block == n2y_block
    assert n2y_block.rich_text[0].mention.block == n2y_block


def test_paragraph_with_child_paragraph():
    parent = mock_block("paragraph", {"rich_text": [mock_rich_text("parent")]}, True)
    children = [mock_paragraph_block([("child", [])])]
    pandoc_ast, markdown = process_parent_block(parent, children)
    assert pandoc_ast == [Para([Str("parent")]), Para([Str("child")])]
    assert markdown == "parent\n\nchild\n"


def test_heading_1():
    notion_block = mock_block("heading_1", {"rich_text": [mock_rich_text("Heading One")]})
    pandoc_ast, markdown = process_block(notion_block)
    assert pandoc_ast == Header(
        1, ("heading-one", [], []), [Str("Heading"), Space(), Str("One")]
    )
    assert markdown == "# Heading One\n"


def test_heading_1_bolding_stripped():
    notion_block = mock_block(
        "heading_1", {"rich_text": [mock_rich_text("Heading One", ["bold"])]}
    )
    pandoc_ast, markdown = process_block(notion_block)
    assert pandoc_ast == Header(
        1, ("heading-one", [], []), [Str("Heading"), Space(), Str("One")]
    )
    assert markdown == "# Heading One\n"


def test_heading_2():
    notion_block = mock_block("heading_2", {"rich_text": [mock_rich_text("Heading Two")]})
    pandoc_ast, markdown = process_block(notion_block)
    assert pandoc_ast == Header(
        2, ("heading-two", [], []), [Str("Heading"), Space(), Str("Two")]
    )
    assert markdown == "## Heading Two\n"


def test_heading_3():
    notion_block = mock_block(
        "heading_3", {"rich_text": [mock_rich_text("Heading Three")]}
    )
    pandoc_ast, markdown = process_block(notion_block)
    assert pandoc_ast == Header(
        3, ("heading-three", [], []), [Str("Heading"), Space(), Str("Three")]
    )
    assert markdown == "### Heading Three\n"


def test_bulleted_list():
    parent = mock_paragraph_block([("Bulleted List", [])], has_children=True)
    children = [
        mock_block("bulleted_list_item", {"rich_text": [mock_rich_text("Item One")]}),
        mock_block("bulleted_list_item", {"rich_text": [mock_rich_text("Item Two")]}),
    ]
    pandoc_ast, markdown = process_parent_block(parent, children)
    assert pandoc_ast == [
        Para([Str("Bulleted"), Space(), Str("List")]),
        BulletList(
            [
                [Plain([Str("Item"), Space(), Str("One")])],
                [Plain([Str("Item"), Space(), Str("Two")])],
            ]
        ),
    ]
    assert markdown == "Bulleted List\n\n-   Item One\n-   Item Two\n"


def test_numbered_list():
    parent = mock_paragraph_block([("Numbered List", [])], has_children=True)
    children = [
        mock_block("numbered_list_item", {"rich_text": [mock_rich_text("Item One")]}),
        mock_block("numbered_list_item", {"rich_text": [mock_rich_text("Item Two")]}),
    ]
    pandoc_ast, markdown = process_parent_block(parent, children)
    assert pandoc_ast == [
        Para([Str("Numbered"), Space(), Str("List")]),
        OrderedList(
            (1, Decimal(), Period()),
            [
                [Plain([Str("Item"), Space(), Str("One")])],
                [Plain([Str("Item"), Space(), Str("Two")])],
            ],
        ),
    ]
    assert markdown == "Numbered List\n\n1.  Item One\n2.  Item Two\n"


def test_page():
    parent = mock_block("child_page", {"title": "Simple Page"}, has_children=True)
    children = [mock_paragraph_block([("Simple page", [])])]
    pandoc_ast, markdown = process_parent_block(parent, children)
    assert pandoc_ast == Pandoc(
        Meta({}),
        [Para([Str("Simple"), Space(), Str("page")])],
    )
    assert markdown == "Simple page\n"


def test_bookmark_with_caption():
    notion_block = mock_block(
        "bookmark",
        {
            "caption": [mock_rich_text("Innolitics")],
            "url": innolitics_website,
        },
    )
    pandoc_ast, markdown = process_block(notion_block)
    assert pandoc_ast == Para(
        [Link(("", [], []), [Str("Innolitics")], (innolitics_website, ""))]
    )
    assert markdown == "[Innolitics](https://innolitics.com)\n"


def test_bookmark_without_caption():
    notion_block = mock_block(
        "bookmark",
        {
            "caption": [],
            "url": innolitics_website,
        },
    )
    pandoc_ast, markdown = process_block(notion_block)
    assert pandoc_ast == Para(
        [
            Link(
                ("", [], []),
                [Str(innolitics_website)],
                (innolitics_website, ""),
            )
        ]
    )
    assert markdown == "<https://innolitics.com>\n"


def test_divider():
    notion_block = mock_block("divider", {"divider": []})
    pandoc_ast, markdown = process_block(notion_block)
    assert pandoc_ast == HorizontalRule()
    assert re.match("-+", markdown)


def test_block_quote():
    notion_block = mock_block(
        "quote",
        {
            "rich_text": [
                mock_rich_text(
                    "In a time of deceit telling the truth is a revolutionary act."
                ),
            ]
        },
    )
    pandoc_ast, markdown = process_block(notion_block)
    assert pandoc_ast == BlockQuote(
        [
            Para(
                [
                    Str("In"),
                    Space(),
                    Str("a"),
                    Space(),
                    Str("time"),
                    Space(),
                    Str("of"),
                    Space(),
                    Str("deceit"),
                    Space(),
                    Str("telling"),
                    Space(),
                    Str("the"),
                    Space(),
                    Str("truth"),
                    Space(),
                    Str("is"),
                    Space(),
                    Str("a"),
                    Space(),
                    Str("revolutionary"),
                    Space(),
                    Str("act."),
                ]
            )
        ]
    )
    expected_markdown = (
        "> In a time of deceit telling the truth is a revolutionary act.\n"
    )
    assert markdown == expected_markdown


@mock.patch.object(Client, "download_file")
def test_image_internal_with_caption(mock_download):
    notion_block = mock_block(
        "image",
        {
            "type": "file",
            "caption": [mock_rich_text("test image")],
            "file": mock_file(example_img),
        },
    )
    mock_download.return_value = "image.png"
    pandoc_ast, markdown = process_block(notion_block)
    assert pandoc_ast == Para(
        [
            Image(
                ("", [], []),
                [Str("test"), Space(), Str("image")],
                ("image.png", "fig:"),
            )
        ]
    )
    assert markdown == "![test image](image.png)\n"


def test_image_external_without_caption():
    notion_block = mock_block(
        "image",
        {
            "type": "external",
            "caption": [],
            "external": {"url": example_img},
        },
    )
    pandoc_ast, markdown = process_block(notion_block)
    assert pandoc_ast == Para(
        [
            Image(
                ("", [], []),
                [],
                (example_img, ""),
            )
        ]
    )
    assert markdown == "![](https://example.com/image.png)\n"


def test_equation_block():
    lhs = "{\\displaystyle i\\hbar {\\frac {d}{dt}}\\vert "
    rhs = "\\Psi (t)\\rangle={\\hat {H}}\\vert \\Psi (t)\\rangle}"
    example_equation = f"{lhs}{rhs}"
    notion_block = mock_block("equation", {"expression": example_equation})
    pandoc_ast, markdown = process_block(notion_block)
    assert pandoc_ast == Para([Math(DisplayMath(), example_equation)])
    expected_lhs = "$${\\displaystyle i\\hbar {\\frac {d}{dt}}\\vert "
    expected_rhs = "\\Psi (t)\\rangle={\\hat {H}}\\vert \\Psi (t)\\rangle}$$\n"
    expected_markdown = f"{expected_lhs}{expected_rhs}"
    assert markdown == expected_markdown


def test_code_block():
    notion_block = mock_block(
        "code",
        {
            "rich_text": [mock_rich_text("const a = 3")],
            "caption": [],
            "language": "javascript",
        },
    )
    pandoc_ast, markdown = process_block(notion_block)
    assert pandoc_ast == CodeBlock(("", ["javascript"], []), "const a = 3")
    assert markdown == "``` javascript\nconst a = 3\n```\n"


def test_table_block():
    parent = mock_block(
        "table",
        {
            "table_width": 2,
            "has_column_header": True,
            "has_row_header": False,
        },
        has_children=True,
    )
    children = [
        mock_block(
            "table_row",
            {
                "cells": [
                    [mock_rich_text("header1")],
                    [mock_rich_text("header2")],
                ]
            },
        ),
        mock_block(
            "table_row",
            {
                "cells": [
                    [mock_rich_text("one")],
                    [mock_rich_text("two")],
                ]
            },
        ),
        mock_block(
            "table_row",
            {
                "cells": [
                    [mock_rich_text("three\n3.5")],
                    [mock_rich_text("four")],
                ]
            },
        ),
    ]
    pandoc_ast, markdown = process_parent_block(parent, children)
    assert pandoc_ast == Table(
        ("", [], []),
        Caption(None, []),
        [(AlignDefault(), ColWidthDefault()), (AlignDefault(), ColWidthDefault())],
        TableHead(
            ("", [], []),
            [
                Row(
                    ("", [], []),
                    [
                        Cell(
                            ("", [], []),
                            AlignDefault(),
                            RowSpan(1),
                            ColSpan(1),
                            [Plain([Str("header1")])],
                        ),
                        Cell(
                            ("", [], []),
                            AlignDefault(),
                            RowSpan(1),
                            ColSpan(1),
                            [Plain([Str("header2")])],
                        ),
                    ],
                )
            ],
        ),
        [
            TableBody(
                ("", [], []),
                RowHeadColumns(0),
                [],
                [
                    Row(
                        ("", [], []),
                        [
                            Cell(
                                ("", [], []),
                                AlignDefault(),
                                RowSpan(1),
                                ColSpan(1),
                                [Plain([Str("one")])],
                            ),
                            Cell(
                                ("", [], []),
                                AlignDefault(),
                                RowSpan(1),
                                ColSpan(1),
                                [Plain([Str("two")])],
                            ),
                        ],
                    ),
                    Row(
                        ("", [], []),
                        [
                            Cell(
                                ("", [], []),
                                AlignDefault(),
                                RowSpan(1),
                                ColSpan(1),
                                [Plain([Str("three"), LineBreak(), Str("3.5")])],
                            ),
                            Cell(
                                ("", [], []),
                                AlignDefault(),
                                RowSpan(1),
                                ColSpan(1),
                                [Plain([Str("four")])],
                            ),
                        ],
                    ),
                ],
            )
        ],
        TableFoot(("", [], []), []),
    )
    assert (
        markdown
        == "  -------------------\n"
        "  header1   header2\n"
        "  --------- ---------\n"
        "  one       two\n"
        "\n  three\\    four\n"
        "  3.5       \n  "
        "-------------------\n"
    )


def test_toggle():
    parent = mock_block(
        "toggle",
        {
            "rich_text": [
                mock_rich_text("Toggle Header"),
            ]
        },
        has_children=True,
    )
    children = [mock_paragraph_block([("Toggle Content", [])])]
    pandoc_ast, markdown = process_parent_block(parent, children)
    assert pandoc_ast == BulletList(
        [
            [
                Para([Str("Toggle"), Space(), Str("Header")]),
                Para([Str("Toggle"), Space(), Str("Content")]),
            ]
        ]
    )
    assert markdown == "-   Toggle Header\n\n    Toggle Content\n"


def test_todo_in_paragraph():
    parent = mock_block(
        "paragraph",
        {
            "rich_text": [
                mock_rich_text("Task List"),
            ]
        },
        has_children=True,
    )
    children = [
        mock_block("to_do", {"rich_text": [mock_rich_text("Task One")], "checked": True}),
        mock_block(
            "to_do", {"rich_text": [mock_rich_text("Task Two")], "checked": False}
        ),
    ]
    pandoc_ast, markdown = process_parent_block(parent, children)
    assert pandoc_ast == [
        Para([Str("Task"), Space(), Str("List")]),
        BulletList(
            [
                [Plain([Str("☒"), Space(), Str("Task"), Space(), Str("One")])],
                [Plain([Str("☐"), Space(), Str("Task"), Space(), Str("Two")])],
            ]
        ),
    ]
    assert markdown == "Task List\n\n-   [x] Task One\n-   [ ] Task Two\n"


@pytest.mark.xfail(reason="Its unclear how to represent empty todos in pandoc")
def test_todo_empty():
    notion_block = mock_block("to_do", {"rich_text": [], "checked": False})
    _, markdown = process_block(notion_block)
    assert markdown == "-   [ ] \n"


def test_callout():
    parent = mock_block(
        "callout", {"rich_text": [mock_rich_text("Callout")]}, has_children=True
    )
    children = [mock_paragraph_block([("Children", [])])]
    pandoc_ast, markdown = process_parent_block(parent, children)
    assert pandoc_ast == [Para([Str("Callout")]), Para([Str("Children")])]
    assert markdown == "Callout\n\nChildren\n"


def test_synced_block_shared():
    original_synced_block = mock_block(
        "synced_block", {"synced_from": None}, has_children=True
    )
    reference_synced_block = mock_block(
        "synced_block",
        {"synced_from": {"type": "block_id", "block_id": "some-block-id"}},
        has_children=True,
    )
    children = [mock_paragraph_block([("synced", [])])]
    original_pandoc_ast, original_markdown = process_parent_block(
        original_synced_block,
        children,
    )
    reference_pandoc_ast, reference_markdown = process_parent_block(
        reference_synced_block,
        children,
    )
    assert original_pandoc_ast == reference_pandoc_ast == [Para([Str("synced")])]
    assert original_markdown == reference_markdown == "synced\n"


def test_synced_block_unshared():
    unshared_reference_synced_block = mock_block(
        "synced_block",
        {"synced_from": {"type": "block_id", "block_id": "some-block-id"}},
        has_children=False,
    )
    unshared_reference_pandoc_ast, unshared_reference_markdown = process_parent_block(
        unshared_reference_synced_block,
        None,
    )
    assert unshared_reference_pandoc_ast is None
    assert unshared_reference_markdown == ""


def test_link_to_page_page():
    mock_link_to_page_block = mock_block(
        "link_to_page",
        {"type": "page_id", "page_id": mock_id()},
    )
    page = mock_page("Linked Page")
    with mock.patch("n2y.notion.Client._get_url", return_value=page):
        pandoc_ast, markdown = process_block(mock_link_to_page_block)
    assert pandoc_ast == Para([Str("Linked"), Space(), Str("Page")])
    assert markdown == "Linked Page\n"


def test_column_block():
    column_block = mock_block("column", {}, has_children=True)
    children = [mock_paragraph_block([("child", [])])]
    with mock.patch.object(
        Client, "get_child_notion_blocks"
    ) as mock_get_child_notion_blocks:
        mock_get_child_notion_blocks.return_value = children
        n2y_block = generate_block(column_block)
    pandoc_ast = n2y_block.to_pandoc()
    assert pandoc_ast == [Para([Str("child")])]


@mock.patch("n2y.notion.Client.get_child_notion_blocks")
def test_column_list_block(mock_get_child_notion_blocks):
    column_list_block = mock_block("column_list", {}, True)
    column1, column2 = mock_block("column", {}, True), mock_block("column", {}, True)
    para1, para2 = mock_paragraph_block([["child1"]]), mock_paragraph_block([["child2"]])
    # Return [column1, column2] for the column list get_child_notion_blocks call
    # and [para1] and [para2] for the get_child_notion_blocks calls of the
    # respective column blocks
    mock_get_child_notion_blocks.side_effect = [[column1, column2], [para1], [para2]]
    pandoc_ast, markdown = process_block(column_list_block)
    assert pandoc_ast == [Para([Str("child1")]), Para([Str("child2")])]
    assert markdown == "child1\n\nchild2\n"


def test_toc_item_block():
    contents = {
        "header": toc_headers[0],
        "subheaders": toc_headers[1:],
        "rich_text": mock_rich_text_array(
            [("Foo Items Header", None, "#foo-items-header")]
        ),
    }
    toc_item_block = mock_block("table_of_contents_item", contents)
    pandoc_ast, markdown = process_block(toc_item_block)
    assert pandoc_ast == toc_item_ast
    assert (
        markdown
        == """\
[Foo Items Header](#foo-items-header)

1.  [Foo: Bar](#foo-bar)
2.  [Foo: Ski](#foo-ski)
    1.  [First Foo Ski](#first-foo-ski)
"""
    )


def test_toc_block():
    toc_block = mock_block("table_of_contents", {})
    client = Client("")
    toc = client.wrap_notion_block(toc_block, None, True)
    toc.render_toc([*toc_headers, *toc_headers])
    pandoc_ast = toc.to_pandoc()
    markdown = pandoc_ast_to_markdown(pandoc_ast)
    assert pandoc_ast == [
        OrderedList((1, Decimal(), Period()), [toc_item_ast, toc_item_ast])
    ]
    assert (
        markdown
        == """\
1.  [Foo Items Header](#foo-items-header)
    1.  [Foo: Bar](#foo-bar)
    2.  [Foo: Ski](#foo-ski)
        1.  [First Foo Ski](#first-foo-ski)
2.  [Foo Items Header](#foo-items-header)
    1.  [Foo: Bar](#foo-bar)
    2.  [Foo: Ski](#foo-ski)
        1.  [First Foo Ski](#first-foo-ski)
"""
    )


def test_toc_block_starting_h2():
    toc_block = mock_block("table_of_contents", {})
    client = Client("")
    toc = client.wrap_notion_block(toc_block, None, True)
    toc.render_toc([*toc_headers_starting_h2, *toc_headers_starting_h2])
    pandoc_ast = toc.to_pandoc()
    markdown = pandoc_ast_to_markdown(pandoc_ast)
    assert (
        markdown
        == """\
1.  [Foo Header H2](#foo-header-h2)
    1.  [Foo: Bar H3](#foo-bar-h3)
    2.  [Foo: Ski H3](#foo-ski-h3)
        1.  [First Foo Ski](#first-foo-ski)
2.  [Foo Header H2](#foo-header-h2)
    1.  [Foo: Bar H3](#foo-bar-h3)
    2.  [Foo: Ski H3](#foo-ski-h3)
        1.  [First Foo Ski](#first-foo-ski)
"""
    )


@mock.patch("n2y.notion.Client.wrap_notion_user")
def test_toc_block_h1_after_h2_base(wrap_notion_user):
    toc_block = mock_block("table_of_contents", {})
    client = Client("")
    wrap_notion_user.return_value = User(client, mock_user())
    page_data = mock_page("Mock Page")
    page = Page(client, page_data)
    toc = client.wrap_notion_block(toc_block, page, True)
    toc.render_toc([*toc_headers_starting_h2, *toc_headers])
    pandoc_ast = toc.to_pandoc()
    markdown = pandoc_ast_to_markdown(pandoc_ast)
    assert (
        markdown
        == """\
1.  [Foo Header H2](#foo-header-h2)
    1.  [Foo: Bar H3](#foo-bar-h3)
    2.  [Foo: Ski H3](#foo-ski-h3)
        1.  [First Foo Ski](#first-foo-ski)
2.  [Foo: Bar](#foo-bar)
3.  [Foo: Ski](#foo-ski)
    1.  [First Foo Ski](#first-foo-ski)
"""
    )



import logging

FORMATTER = logging.Formatter(
    "%(asctime)s - %(levelname)s - n2y.%(module)s: %(message)s"
)
HANDLER = logging.StreamHandler()
HANDLER.setFormatter(FORMATTER)
logging.basicConfig(level=logging.INFO, handlers=[HANDLER])
logger: logging.Logger = logging.getLogger(__name__)


def update_logger(new_logger: logging.Logger):
    global logger
    logger = new_logger
